# DOCS_REF:
# - docs/user/README.md
# - docs/user/04-services-sources-topics.md
# - docs/user/reference/api-surface.md
# DOCS_SYNC:
# - If TopicStream operators or subscribe semantics change, update the
#   referenced docs and examples/docs/{services_sources_topics_example.py,cookbook_topic_flow_topic_example.py}.

from __future__ import annotations

from flownet.api.datastream import DataStream, PartitionedStream
import flownet.api.node_control as node_control_api
from flownet.api.services import get_services
from flownet.compiler.flowtask import FlowTask
from flownet.core.exceptions import FlowDefinitionError


class TopicStream:
    _SERVICE_TOPIC_PREFIX = "service:"

    def __init__(self, topic_id: str):
        self.topic_id = topic_id
        self._flowtask = FlowTask()
        self._head = DataStream(self._flowtask, transformation=None, init=True)
        self._tail: DataStream = self._head
        self._subscribed = False
        self._subscription_id: str | None = None

    def map(self, fn_object) -> "TopicStream":
        self._tail = self._tail.map(fn_object)
        return self

    def pipe(self, fn_object) -> "TopicStream":
        self._tail = self._tail.pipe(fn_object)
        return self

    def filter(self, fn_object) -> "TopicStream":
        self._tail = self._tail.filter(fn_object)
        return self

    def flatmap(self, fn_object) -> "TopicStream":
        self._tail = self._tail.flatmap(fn_object)
        return self

    def tap(self, fn_object) -> "TopicStream":
        self._tail = self._tail.tap(fn_object)
        return self

    def write(self, sink) -> "TopicStream":
        self._tail = self._tail.write(sink)
        return self

    def publish(self, sink) -> "TopicStream":
        return self.write(sink)

    def shuffle(self, **kwargs) -> "TopicStream":
        self._tail = self._tail.shuffle(**kwargs)
        return self

    def key_by(self, **kwargs) -> "TopicStream":
        self._tail = self._tail.key_by(**kwargs)
        return self

    def repartition(self, num_partitions: int, **kwargs) -> "TopicStream":
        self._tail = self._tail.repartition(num_partitions=num_partitions, **kwargs)
        return self

    def reduce(self, **kwargs) -> "TopicStream":
        if not isinstance(self._tail, PartitionedStream):
            raise FlowDefinitionError("reduce() requires shuffle() or repartition() first.")
        self._tail = self._tail.reduce(**kwargs)
        return self

    def process_with_state(self, handler_ref, **kwargs) -> "TopicStream":
        if not isinstance(self._tail, PartitionedStream):
            raise FlowDefinitionError("process_with_state() requires shuffle(), key_by(), or repartition() first.")
        self._tail = self._tail.process_with_state(handler_ref, **kwargs)
        return self

    def set_return(self) -> "TopicStream":
        self._tail.set_return()
        return self

    # Backward-compatible alias (same as return marker).
    def set_sink(self) -> "TopicStream":
        return self.set_return()

    def _service_name(self) -> str | None:
        if self.topic_id.startswith(self._SERVICE_TOPIC_PREFIX):
            return self.topic_id[len(self._SERVICE_TOPIC_PREFIX):]
        return None

    def subscribe(self, source_address: str | None = None) -> "TopicStream":
        service_name = self._service_name()
        if source_address is None and service_name is None:
            self._subscribe_topic_registry()
            return self

        if not self._flowtask.pipeline:
            raise ValueError("TopicStream has no operators. Add map/pipe/write before subscribe().")

        hub = get_services()
        topics = hub.require_topics()
        execution_address = hub.local_address()

        if service_name is None:
            raise ValueError("source_address subscribe is only supported for service topics.")

        if source_address is None or source_address == execution_address:
            if topics.topic_registry is not None:
                self._subscribe_topic_registry()
                return self
            if not self._subscribed:
                self._subscription_id = topics.require_service_registry().subscribe_flow(
                    service_name,
                    self._flowtask,
                    execution_address,
                )
                self._subscribed = True
            return self

        self._subscription_id = node_control_api.subscribe_service_flow_on(
            source_address,
            service_name,
            self._flowtask,
            execution_address,
        )
        self._subscribed = True
        return self

    def submit(self, source_address: str | None = None) -> "TopicStream":
        return self.subscribe(source_address=source_address)

    def unsubscribe(self, source_address: str | None = None) -> bool:
        if not self._subscription_id:
            return False

        service_name = self._service_name()
        if source_address is None and service_name is None:
            return self._unsubscribe_topic_registry()

        hub = get_services()
        topics = hub.require_topics()
        execution_address = hub.local_address()

        if service_name is None:
            raise ValueError("source_address unsubscribe is only supported for service topics.")

        if source_address is None or source_address == execution_address:
            if topics.topic_registry is not None:
                return self._unsubscribe_topic_registry()
            return topics.require_service_registry().unsubscribe_flow(self._subscription_id)

        return node_control_api.unsubscribe_service_flow_on(
            source_address,
            self._subscription_id,
        )

    def _subscribe_topic_registry(self) -> None:
        if not self._flowtask.pipeline:
            raise ValueError("TopicStream has no operators. Add map/pipe/write before subscribe().")
        hub = get_services()
        registry = hub.require_topics().require_topic_registry()
        if not self._subscribed:
            self._subscription_id = registry.attach_task(
                self.topic_id,
                self._flowtask,
                hub.local_address(),
            )
            self._subscribed = True

    def _unsubscribe_topic_registry(self) -> bool:
        hub = get_services()
        registry = hub.require_topics().require_topic_registry()
        return registry.detach_task(self._subscription_id)


def topic_stream(topic_id: str) -> TopicStream:
    return TopicStream(topic_id)


def from_topic(topic_id: str) -> TopicStream:
    return topic_stream(topic_id)


def from_service(service_name: str) -> TopicStream:
    normalized_name = _normalize_service_name(service_name)
    return from_topic(f"{TopicStream._SERVICE_TOPIC_PREFIX}{normalized_name}")


def _normalize_service_name(service_name: str) -> str:
    if not isinstance(service_name, str):
        raise TypeError("service_name must be a string.")

    normalized = service_name.strip()
    if not normalized:
        raise ValueError("service_name must be a non-empty string.")
    if normalized.startswith(TopicStream._SERVICE_TOPIC_PREFIX):
        raise ValueError("from_service() expects service name without the 'service:' prefix.")
    return normalized


__all__ = [
    "TopicStream",
    "topic_stream",
    "from_topic",
    "from_service",
]



from __future__ import annotations

from collections import defaultdict, deque
import threading
import time
from dataclasses import dataclass, replace
import math
from typing import Callable, Iterable, Any

from flownet.api.scheduler import SchedulePlan, SimpleScheduler
from flownet.runtime.cluster_view import ClusterNode, ClusterView


DEFAULT_LOCALITY_ORDER = ("local", "zone", "region", "remote")
DEFAULT_RESERVATION_TIMEOUT_MS = 30000
DEFAULT_LATENCY_SAMPLE_WINDOW = 2048


@dataclass(frozen=True)
class PlanResolutionRequest:
    stage_id: str
    request_id: str
    num_partitions: int
    affinity_hint: str | None = None
    addresses: tuple[str, ...] | None = None
    merge_addresses: tuple[str, ...] | None = None
    merge_partitions: dict[str, str] | None = None
    include_merge_address: bool = True
    merge_sharded: bool = False
    epoch: int = 1
    plan_id: str | None = None
    require_healthy: bool = True
    locality_order: tuple[str, ...] = DEFAULT_LOCALITY_ORDER
    resource_requirements: dict[str, int | float] | None = None
    allow_resource_fallback: bool = True
    lease_owner_kind: str | None = None
    lease_owner_id: str | None = None
    reservation_timeout_ms: int | None = None
    autoscale_min_workers: int | None = None
    autoscale_max_workers: int | None = None
    autoscale_queue_depth: int | None = None
    autoscale_target_queue_per_worker: int | None = None
    autoscale_observed_latency_ms: float | None = None
    autoscale_latency_budget_ms: float | None = None
    autoscale_observed_throughput_rps: float | None = None
    autoscale_target_throughput_rps: float | None = None
    autoscale_cooldown_ms: int | None = None
    spillover_stage1_tiers: tuple[str, ...] = ("local", "zone")
    spillover_stage2_tiers: tuple[str, ...] = ("region",)
    spillover_allow_remote: bool = True
    spillover_max_remote_nodes: int | None = None
    spillover_ttl_ms: int | None = None
    spillover_claim_id: str | None = None


@dataclass(frozen=True)
class PlanResolution:
    plan: SchedulePlan
    addresses: tuple[str, ...]
    merge_addresses: tuple[str, ...]
    strategy: str
    resource_fallback_used: bool = False


@dataclass
class _AutoscaleStageState:
    desired_workers: int
    cooldown_ms: int
    last_scale_ts: float = 0.0
    last_direction: str = "steady"
    last_reason: str = "steady"
    last_decision_ts: float = 0.0
    decisions: int = 0
    cooldown_suppressed: int = 0


class ClusterPlanResolver:
    """Resolve scheduler plans from ClusterView using locality + health ranking.

    Design guardrail:
    - Resolver consumes ClusterView snapshots and emits placement plans.
    - Resolver must not mutate/own membership lifecycle; ClusterView maintenance
      stays in runtime cluster/gossip paths.
    """

    def __init__(
        self,
        cluster_view: ClusterView,
        *,
        time_fn: Callable[[], float] | None = None,
    ):
        self._cluster_view = cluster_view
        self._time_fn = time_fn or time.time
        self._metrics_lock = threading.Lock()
        self._plan_resolution_count = 0
        self._resource_fallback_count = 0
        self._locality_hits: dict[str, int] = {
            "local": 0,
            "zone": 0,
            "region": 0,
            "remote": 0,
            "unknown": 0,
        }
        self._locality_total_hits = 0
        self._staleness_count = 0
        self._staleness_ms_sum = 0.0
        self._staleness_ms_max = 0.0
        self._staleness_ms_latest = 0.0
        self._latency_count = 0
        self._latency_ms_sum = 0.0
        self._latency_ms_min: float | None = None
        self._latency_ms_max = 0.0
        self._latency_ms_recent: deque[float] = deque(maxlen=DEFAULT_LATENCY_SAMPLE_WINDOW)
        # Owner-aware fairness state (best-effort in-memory lease reservations).
        self._owner_resolution_count = 0
        self._reservation_entries: list[tuple[str, str, float]] = []
        self._owner_active_reservations: dict[str, int] = defaultdict(int)
        self._node_active_reservations: dict[str, int] = defaultdict(int)
        self._owner_node_active_reservations: dict[tuple[str, str], int] = defaultdict(int)
        # Autoscaling control-loop state (best-effort, stage-local).
        self._autoscale_states: dict[str, _AutoscaleStageState] = {}
        self._autoscale_decision_count = 0
        self._autoscale_scale_up_count = 0
        self._autoscale_scale_down_count = 0
        self._autoscale_cooldown_suppressed_count = 0
        # Spillover control state (best-effort, in-memory).
        self._spillover_decision_count = 0
        self._spillover_remote_used_count = 0
        self._spillover_remote_blocked_count = 0
        self._spillover_claim_created_count = 0
        self._spillover_claim_reused_count = 0
        self._spillover_claim_expiry: dict[str, float] = {}
        # Request idempotence for owner lease reservations.
        self._owner_request_claim_expiry: dict[str, float] = {}

    def resolve(self, request: PlanResolutionRequest) -> PlanResolution:
        started_at = self._time_fn()
        self._prune_expired_reservations(now=started_at)
        effective_request, autoscale_applied = self._apply_autoscale(
            request=request,
            now=started_at,
        )
        addresses, resource_fallback_used, spillover_strategy = self._resolve_addresses(
            effective_request,
            now=started_at,
        )
        merge_addresses = tuple(effective_request.merge_addresses or addresses)
        plan = SimpleScheduler(addresses).make_plan(
            stage_id=effective_request.stage_id,
            request_id=effective_request.request_id,
            num_partitions=effective_request.num_partitions,
            merge_addresses=merge_addresses,
            merge_partitions=effective_request.merge_partitions,
            include_merge_address=effective_request.include_merge_address,
            merge_sharded=effective_request.merge_sharded,
            epoch=effective_request.epoch,
            plan_id=effective_request.plan_id,
        )
        strategy = "explicit_addresses" if request.addresses else "cluster_view_locality_health"
        if resource_fallback_used:
            strategy = f"{strategy}+resource_fallback"
        if autoscale_applied:
            strategy = f"{strategy}+autoscale"
        if spillover_strategy:
            strategy = f"{strategy}+{spillover_strategy}"
        completed_at = self._time_fn()
        self._record_owner_reservations(
            request=effective_request,
            plan=plan,
            now=completed_at,
        )
        self._record_metrics(
            latency_ms=(completed_at - started_at) * 1000.0,
            staleness_ms=self._cluster_view.estimate_staleness_ms(now=completed_at),
            resource_fallback_used=resource_fallback_used,
            locality_hits=self._count_locality_hits(plan),
        )
        return PlanResolution(
            plan=plan,
            addresses=addresses,
            merge_addresses=merge_addresses,
            strategy=strategy,
            resource_fallback_used=resource_fallback_used,
        )

    def resolve_plan(
        self,
        *,
        stage_id: str,
        request_id: str,
        num_partitions: int,
        affinity_hint: str | None = None,
        addresses: Iterable[str] | None = None,
        merge_addresses: Iterable[str] | None = None,
        merge_partitions: dict[str, str] | None = None,
        include_merge_address: bool = True,
        merge_sharded: bool = False,
        epoch: int = 1,
        plan_id: str | None = None,
        require_healthy: bool = True,
        locality_order: tuple[str, ...] = DEFAULT_LOCALITY_ORDER,
        resource_requirements: dict[str, int | float] | None = None,
        allow_resource_fallback: bool = True,
        lease_owner_kind: str | None = None,
        lease_owner_id: str | None = None,
        reservation_timeout_ms: int | None = None,
        autoscale_min_workers: int | None = None,
        autoscale_max_workers: int | None = None,
        autoscale_queue_depth: int | None = None,
        autoscale_target_queue_per_worker: int | None = None,
        autoscale_observed_latency_ms: float | None = None,
        autoscale_latency_budget_ms: float | None = None,
        autoscale_observed_throughput_rps: float | None = None,
        autoscale_target_throughput_rps: float | None = None,
        autoscale_cooldown_ms: int | None = None,
        spillover_stage1_tiers: tuple[str, ...] = ("local", "zone"),
        spillover_stage2_tiers: tuple[str, ...] = ("region",),
        spillover_allow_remote: bool = True,
        spillover_max_remote_nodes: int | None = None,
        spillover_ttl_ms: int | None = None,
        spillover_claim_id: str | None = None,
    ) -> SchedulePlan:
        request = PlanResolutionRequest(
            stage_id=stage_id,
            request_id=request_id,
            num_partitions=num_partitions,
            affinity_hint=affinity_hint,
            addresses=tuple(addresses) if addresses is not None else None,
            merge_addresses=tuple(merge_addresses) if merge_addresses is not None else None,
            merge_partitions=merge_partitions,
            include_merge_address=include_merge_address,
            merge_sharded=merge_sharded,
            epoch=epoch,
            plan_id=plan_id,
            require_healthy=require_healthy,
            locality_order=locality_order,
            resource_requirements=resource_requirements,
            allow_resource_fallback=allow_resource_fallback,
            lease_owner_kind=lease_owner_kind,
            lease_owner_id=lease_owner_id,
            reservation_timeout_ms=reservation_timeout_ms,
            autoscale_min_workers=autoscale_min_workers,
            autoscale_max_workers=autoscale_max_workers,
            autoscale_queue_depth=autoscale_queue_depth,
            autoscale_target_queue_per_worker=autoscale_target_queue_per_worker,
            autoscale_observed_latency_ms=autoscale_observed_latency_ms,
            autoscale_latency_budget_ms=autoscale_latency_budget_ms,
            autoscale_observed_throughput_rps=autoscale_observed_throughput_rps,
            autoscale_target_throughput_rps=autoscale_target_throughput_rps,
            autoscale_cooldown_ms=autoscale_cooldown_ms,
            spillover_stage1_tiers=spillover_stage1_tiers,
            spillover_stage2_tiers=spillover_stage2_tiers,
            spillover_allow_remote=spillover_allow_remote,
            spillover_max_remote_nodes=spillover_max_remote_nodes,
            spillover_ttl_ms=spillover_ttl_ms,
            spillover_claim_id=spillover_claim_id,
        )
        return self.resolve(request).plan

    def metrics_snapshot(self) -> dict[str, object]:
        now = self._time_fn()
        with self._metrics_lock:
            self._prune_expired_locked(now=now)
            total = self._plan_resolution_count
            fallback_rate = (
                float(self._resource_fallback_count) / float(total) if total else 0.0
            )
            hit_total = self._locality_total_hits
            locality_rate = {
                tier: (float(count) / float(hit_total) if hit_total else 0.0)
                for tier, count in self._locality_hits.items()
            }
            avg_staleness = (
                self._staleness_ms_sum / float(self._staleness_count)
                if self._staleness_count
                else 0.0
            )
            avg_latency = (
                self._latency_ms_sum / float(self._latency_count)
                if self._latency_count
                else 0.0
            )
            owner_active = dict(self._owner_active_reservations)
            owner_counts = list(owner_active.values())
            max_owner_active = max(owner_counts) if owner_counts else 0
            min_owner_active = min(owner_counts) if owner_counts else 0
            return {
                "cluster_view_staleness_ms": {
                    "latest": self._staleness_ms_latest,
                    "avg": avg_staleness,
                    "max": self._staleness_ms_max,
                    "samples": self._staleness_count,
                },
                "scheduler_resource_fallback_rate": {
                    "rate": fallback_rate,
                    "fallback_count": self._resource_fallback_count,
                    "total": total,
                },
                "scheduler_locality_tier_hit_rate": {
                    "rate": locality_rate,
                    "counts": dict(self._locality_hits),
                    "total_hits": hit_total,
                },
                "plan_resolution_latency_ms": {
                    "avg": avg_latency,
                    "min": 0.0 if self._latency_ms_min is None else self._latency_ms_min,
                    "max": self._latency_ms_max,
                    "samples": self._latency_count,
                    "p50": _quantile(self._latency_ms_recent, 0.50),
                    "p95": _quantile(self._latency_ms_recent, 0.95),
                    "quantile_window_samples": len(self._latency_ms_recent),
                },
                "scheduler_owner_fairness": {
                    "owner_resolution_count": self._owner_resolution_count,
                    "active_owners": len(owner_active),
                    "active_reservations": len(self._reservation_entries),
                    "max_owner_active_reservations": max_owner_active,
                    "min_owner_active_reservations": min_owner_active,
                    "owner_active_reservations": owner_active,
                    "node_active_reservations": dict(self._node_active_reservations),
                },
                "scheduler_spillover": {
                    "decision_count": self._spillover_decision_count,
                    "remote_used_count": self._spillover_remote_used_count,
                    "remote_blocked_count": self._spillover_remote_blocked_count,
                    "claim_created_count": self._spillover_claim_created_count,
                    "claim_reused_count": self._spillover_claim_reused_count,
                    "active_claims": len(self._spillover_claim_expiry),
                },
                "scheduler_autoscaling": self._autoscale_snapshot_locked(),
            }

    def reset_metrics(self) -> None:
        with self._metrics_lock:
            self._plan_resolution_count = 0
            self._resource_fallback_count = 0
            for key in list(self._locality_hits.keys()):
                self._locality_hits[key] = 0
            self._locality_total_hits = 0
            self._staleness_count = 0
            self._staleness_ms_sum = 0.0
            self._staleness_ms_max = 0.0
            self._staleness_ms_latest = 0.0
            self._latency_count = 0
            self._latency_ms_sum = 0.0
            self._latency_ms_min = None
            self._latency_ms_max = 0.0
            self._latency_ms_recent.clear()
            self._owner_resolution_count = 0
            self._reservation_entries = []
            self._owner_active_reservations.clear()
            self._node_active_reservations.clear()
            self._owner_node_active_reservations.clear()
            self._autoscale_states.clear()
            self._autoscale_decision_count = 0
            self._autoscale_scale_up_count = 0
            self._autoscale_scale_down_count = 0
            self._autoscale_cooldown_suppressed_count = 0
            self._spillover_decision_count = 0
            self._spillover_remote_used_count = 0
            self._spillover_remote_blocked_count = 0
            self._spillover_claim_created_count = 0
            self._spillover_claim_reused_count = 0
            self._spillover_claim_expiry.clear()
            self._owner_request_claim_expiry.clear()

    def _resolve_addresses(
        self,
        request: PlanResolutionRequest,
        *,
        now: float,
    ) -> tuple[tuple[str, ...], bool, str | None]:
        if request.addresses:
            addresses = _dedupe_preserve_order(request.addresses)
            if addresses:
                return addresses, False, None
            raise ValueError("PlanResolutionRequest.addresses must not be empty.")

        ranked, resource_fallback_used, spillover_strategy = self._ranked_nodes(request, now=now)
        if request.affinity_hint:
            hint_node = self._cluster_view.resolve_node_for_key(
                request.stage_id,
                affinity_hint=request.affinity_hint,
            )
            ranked = _promote_node(ranked, hint_node.node_id)
        addresses = _dedupe_preserve_order(node.address for node in ranked)
        if not addresses:
            raise ValueError("No schedulable nodes available from ClusterView.")
        return addresses, resource_fallback_used, spillover_strategy

    def _ranked_nodes(
        self,
        request: PlanResolutionRequest,
        *,
        now: float,
    ) -> tuple[list[ClusterNode], bool, str | None]:
        nodes = [
            node
            for node in self._cluster_view.list_nodes()
            if self._cluster_view.is_node_schedulable(node.node_id)
        ]
        if request.require_healthy:
            healthy_nodes = [
                node for node in nodes if self._cluster_view.is_node_healthy(node.node_id)
            ]
            if healthy_nodes:
                nodes = healthy_nodes
        resource_fallback_used = False
        if request.resource_requirements:
            matched = [
                node
                for node in nodes
                if _resource_match(node.resource_summary, request.resource_requirements)
            ]
            if matched:
                nodes = matched
            elif not request.allow_resource_fallback:
                raise ValueError(
                    "No nodes satisfy resource_requirements and allow_resource_fallback=False."
                )
            else:
                resource_fallback_used = True

        order = {tier: idx for idx, tier in enumerate(request.locality_order)}
        max_priority = len(order)
        owner_key = _owner_key(request)
        owner_node_counts: dict[str, int] = {}
        node_counts: dict[str, int] = {}
        if owner_key is not None:
            with self._metrics_lock:
                owner_node_counts = {
                    node_id: count
                    for (entry_owner, node_id), count in self._owner_node_active_reservations.items()
                    if entry_owner == owner_key
                }
                node_counts = dict(self._node_active_reservations)

        def sort_key(node: ClusterNode):
            tier = self._cluster_view.locality_tier(node.node_id)
            priority = order.get(tier, max_priority)
            freshness = 0.0 if node.last_seen is None else node.last_seen
            if owner_key is None:
                return (priority, -freshness, node.node_id)
            owner_load = owner_node_counts.get(node.node_id, 0)
            global_load = node_counts.get(node.node_id, 0)
            return (priority, owner_load, global_load, -freshness, node.node_id)

        ranked = sorted(nodes, key=sort_key)
        ranked, spillover_strategy = self._apply_spillover_policy(
            ranked=ranked,
            request=request,
            now=now,
        )
        return ranked, resource_fallback_used, spillover_strategy

    def _record_owner_reservations(
        self,
        *,
        request: PlanResolutionRequest,
        plan: SchedulePlan,
        now: float,
    ) -> None:
        owner_key = _owner_key(request)
        if owner_key is None:
            return

        ttl_ms = request.reservation_timeout_ms
        if ttl_ms is None:
            ttl_ms = DEFAULT_RESERVATION_TIMEOUT_MS
        try:
            ttl_ms_int = int(ttl_ms)
        except (TypeError, ValueError) as exc:
            raise ValueError("reservation_timeout_ms must be an integer > 0.") from exc
        if ttl_ms_int <= 0:
            raise ValueError("reservation_timeout_ms must be an integer > 0.")

        expires_at = now + (float(ttl_ms_int) / 1000.0)
        node_id_by_address = {node.address: node.node_id for node in self._cluster_view.list_nodes()}

        with self._metrics_lock:
            self._prune_expired_locked(now=now)
            request_key = f"{owner_key}:{request.stage_id}:{request.request_id}"
            prev_claim_expiry = float(self._owner_request_claim_expiry.get(request_key, 0.0))
            if prev_claim_expiry > now:
                return
            self._owner_resolution_count += 1
            for address in plan.partitions.values():
                node_id = node_id_by_address.get(address, f"addr:{address}")
                self._reservation_entries.append((owner_key, node_id, expires_at))
                self._owner_active_reservations[owner_key] += 1
                self._node_active_reservations[node_id] += 1
                self._owner_node_active_reservations[(owner_key, node_id)] += 1
            self._owner_request_claim_expiry[request_key] = expires_at

    def _prune_expired_reservations(self, *, now: float) -> None:
        with self._metrics_lock:
            self._prune_expired_locked(now=now)

    def _prune_expired_locked(self, *, now: float) -> None:
        expired_spillover_claims = [
            key
            for key, expires_at in self._spillover_claim_expiry.items()
            if expires_at <= now
        ]
        for key in expired_spillover_claims:
            self._spillover_claim_expiry.pop(key, None)

        expired_owner_claims = [
            key
            for key, expires_at in self._owner_request_claim_expiry.items()
            if expires_at <= now
        ]
        for key in expired_owner_claims:
            self._owner_request_claim_expiry.pop(key, None)

        if not self._reservation_entries:
            return

        alive_entries: list[tuple[str, str, float]] = []
        owner_active: dict[str, int] = defaultdict(int)
        node_active: dict[str, int] = defaultdict(int)
        owner_node_active: dict[tuple[str, str], int] = defaultdict(int)

        for owner_key, node_id, expires_at in self._reservation_entries:
            if expires_at <= now:
                continue
            alive_entries.append((owner_key, node_id, expires_at))
            owner_active[owner_key] += 1
            node_active[node_id] += 1
            owner_node_active[(owner_key, node_id)] += 1

        self._reservation_entries = alive_entries
        self._owner_active_reservations = owner_active
        self._node_active_reservations = node_active
        self._owner_node_active_reservations = owner_node_active

    def _apply_spillover_policy(
        self,
        *,
        ranked: list[ClusterNode],
        request: PlanResolutionRequest,
        now: float,
    ) -> tuple[list[ClusterNode], str | None]:
        if not ranked:
            return ranked, None
        if not _spillover_policy_enabled(request):
            return ranked, None

        target_nodes = min(max(1, int(request.num_partitions)), len(ranked))
        stage1 = set(request.spillover_stage1_tiers)
        stage2 = set(request.spillover_stage2_tiers)
        selected: list[ClusterNode] = []
        selected_ids: set[str] = set()

        def _append_from_tiers(tiers: set[str]) -> None:
            if len(selected) >= target_nodes:
                return
            for node in ranked:
                if len(selected) >= target_nodes:
                    return
                if node.node_id in selected_ids:
                    continue
                tier = self._cluster_view.locality_tier(node.node_id)
                if tier not in tiers:
                    continue
                selected.append(node)
                selected_ids.add(node.node_id)

        _append_from_tiers(stage1)
        _append_from_tiers(stage2)

        strategy_parts: list[str] = []
        remote_needed = len(selected) < target_nodes
        remote_used = False
        remote_blocked = False
        claim_created = False
        claim_reused = False
        if remote_needed:
            remote_pool = [
                node
                for node in ranked
                if node.node_id not in selected_ids
                and self._cluster_view.locality_tier(node.node_id) not in stage1.union(stage2)
            ]
            allow_remote = request.spillover_allow_remote
            if allow_remote and request.spillover_ttl_ms is not None:
                allow_remote, claim_created, claim_reused = self._acquire_spillover_claim(
                    request=request,
                    now=now,
                )
            if allow_remote:
                max_remote = request.spillover_max_remote_nodes
                if max_remote is not None:
                    try:
                        max_remote = max(0, int(max_remote))
                    except (TypeError, ValueError):
                        max_remote = 0
                remaining = target_nodes - len(selected)
                allowed = len(remote_pool)
                if max_remote is not None:
                    allowed = min(allowed, max_remote)
                allowed = min(allowed, remaining)
                for node in remote_pool[:allowed]:
                    selected.append(node)
                    selected_ids.add(node.node_id)
                remote_used = allowed > 0
                remote_blocked = remaining > allowed
            else:
                remote_blocked = True

        if not selected:
            return ranked, None

        if remote_used:
            strategy_parts.append("spillover_remote")
        if remote_blocked:
            strategy_parts.append("spillover_limited")
        with self._metrics_lock:
            self._prune_expired_locked(now=now)
            self._spillover_decision_count += 1
            if remote_used:
                self._spillover_remote_used_count += 1
            if remote_blocked:
                self._spillover_remote_blocked_count += 1
            if claim_created:
                self._spillover_claim_created_count += 1
            if claim_reused:
                self._spillover_claim_reused_count += 1
        return selected, "+".join(strategy_parts) if strategy_parts else "spillover_local_first"

    def _acquire_spillover_claim(
        self,
        *,
        request: PlanResolutionRequest,
        now: float,
    ) -> tuple[bool, bool, bool]:
        ttl_ms = request.spillover_ttl_ms
        if ttl_ms is None:
            return True, False, False
        try:
            ttl_ms_int = int(ttl_ms)
        except (TypeError, ValueError):
            return False, False, False
        if ttl_ms_int <= 0:
            return False, False, False

        claim_key = _spillover_claim_key(request)
        expires_at = now + (float(ttl_ms_int) / 1000.0)
        with self._metrics_lock:
            self._prune_expired_locked(now=now)
            prev = float(self._spillover_claim_expiry.get(claim_key, 0.0))
            if prev > now:
                return True, False, True
            self._spillover_claim_expiry[claim_key] = expires_at
        return True, True, False

    def _apply_autoscale(
        self,
        *,
        request: PlanResolutionRequest,
        now: float,
    ) -> tuple[PlanResolutionRequest, bool]:
        enabled = (
            request.autoscale_min_workers is not None
            or request.autoscale_max_workers is not None
            or request.autoscale_queue_depth is not None
            or request.autoscale_observed_latency_ms is not None
            or request.autoscale_observed_throughput_rps is not None
        )
        if not enabled:
            return request, False

        min_workers = request.autoscale_min_workers
        if min_workers is None:
            min_workers = 1
        max_workers = request.autoscale_max_workers
        if max_workers is None:
            max_workers = max(min_workers, request.num_partitions)
        min_workers = int(min_workers)
        max_workers = int(max_workers)
        if min_workers <= 0:
            raise ValueError("autoscale_min_workers must be > 0.")
        if max_workers < min_workers:
            raise ValueError("autoscale_max_workers must be >= autoscale_min_workers.")

        queue_target_size = request.autoscale_target_queue_per_worker
        if queue_target_size is None:
            queue_target_size = 32
        queue_target_size = int(queue_target_size)
        if queue_target_size <= 0:
            raise ValueError("autoscale_target_queue_per_worker must be > 0.")

        cooldown_ms = request.autoscale_cooldown_ms
        if cooldown_ms is None:
            cooldown_ms = 5000
        cooldown_ms = int(cooldown_ms)
        if cooldown_ms < 0:
            raise ValueError("autoscale_cooldown_ms must be >= 0.")

        stage_key = request.stage_id
        with self._metrics_lock:
            state = self._autoscale_states.get(stage_key)
            if state is None:
                initial = _clamp_int(request.num_partitions, min_workers, max_workers)
                state = _AutoscaleStageState(
                    desired_workers=initial,
                    cooldown_ms=cooldown_ms,
                )
                self._autoscale_states[stage_key] = state
            else:
                state.cooldown_ms = cooldown_ms

            current = _clamp_int(state.desired_workers, min_workers, max_workers)
            candidate = current
            reasons: list[str] = []

            if request.autoscale_queue_depth is not None:
                queue_depth = int(request.autoscale_queue_depth)
                if queue_depth < 0:
                    queue_depth = 0
                queue_target = int(math.ceil(float(queue_depth) / float(queue_target_size)))
                queue_target = _clamp_int(queue_target, min_workers, max_workers)
                if queue_target != candidate:
                    candidate = queue_target
                reasons.append("queue_depth")

            observed_latency = request.autoscale_observed_latency_ms
            latency_budget = request.autoscale_latency_budget_ms
            if (
                observed_latency is not None
                and latency_budget is not None
                and float(observed_latency) > float(latency_budget)
                and candidate < max_workers
            ):
                candidate = min(candidate + 1, max_workers)
                reasons.append("latency_pressure")

            observed_tput = request.autoscale_observed_throughput_rps
            target_tput = request.autoscale_target_throughput_rps
            if observed_tput is not None and target_tput is not None:
                observed_tput = float(observed_tput)
                target_tput = float(target_tput)
                queue_depth = int(request.autoscale_queue_depth or 0)
                if observed_tput < (target_tput * 0.9) and candidate < max_workers:
                    candidate = min(candidate + 1, max_workers)
                    reasons.append("throughput_deficit")
                elif (
                    observed_tput > (target_tput * 1.3)
                    and queue_depth == 0
                    and candidate > min_workers
                ):
                    candidate = max(candidate - 1, min_workers)
                    reasons.append("throughput_surplus")

            direction = "steady"
            if candidate > current:
                direction = "scale_up"
            elif candidate < current:
                direction = "scale_down"

            if direction != "steady" and cooldown_ms > 0:
                cooldown_s = float(cooldown_ms) / 1000.0
                if now < (state.last_scale_ts + cooldown_s):
                    candidate = current
                    direction = "steady"
                    reasons = ["cooldown_suppressed", *reasons]
                    state.cooldown_suppressed += 1
                    self._autoscale_cooldown_suppressed_count += 1

            if direction == "scale_up":
                self._autoscale_scale_up_count += 1
                state.last_scale_ts = now
            elif direction == "scale_down":
                self._autoscale_scale_down_count += 1
                state.last_scale_ts = now

            state.desired_workers = candidate
            state.last_direction = direction
            state.last_reason = ",".join(reasons) if reasons else "steady"
            state.last_decision_ts = now
            state.decisions += 1
            self._autoscale_decision_count += 1

            return replace(request, num_partitions=candidate), candidate != request.num_partitions

    def _autoscale_snapshot_locked(self) -> dict[str, object]:
        stage_snapshot: dict[str, dict[str, object]] = {}
        for stage_id, state in self._autoscale_states.items():
            stage_snapshot[stage_id] = {
                "desired_workers": state.desired_workers,
                "last_direction": state.last_direction,
                "last_reason": state.last_reason,
                "last_decision_ts": state.last_decision_ts,
                "cooldown_ms": state.cooldown_ms,
                "decisions": state.decisions,
                "cooldown_suppressed": state.cooldown_suppressed,
            }
        return {
            "decision_count": self._autoscale_decision_count,
            "scale_up_count": self._autoscale_scale_up_count,
            "scale_down_count": self._autoscale_scale_down_count,
            "cooldown_suppressed_count": self._autoscale_cooldown_suppressed_count,
            "active_stages": len(stage_snapshot),
            "stages": stage_snapshot,
        }

    def _count_locality_hits(self, plan: SchedulePlan) -> dict[str, int]:
        tier_by_address = self._tier_by_address()
        result: dict[str, int] = {}
        for address in plan.partitions.values():
            tier = tier_by_address.get(address, "unknown")
            result[tier] = result.get(tier, 0) + 1
        return result

    def _tier_by_address(self) -> dict[str, str]:
        mapping: dict[str, str] = {}
        for node in self._cluster_view.list_nodes():
            mapping.setdefault(node.address, self._cluster_view.locality_tier(node.node_id))
        return mapping

    def _record_metrics(
        self,
        *,
        latency_ms: float,
        staleness_ms: float,
        resource_fallback_used: bool,
        locality_hits: dict[str, int],
    ) -> None:
        safe_latency = max(0.0, latency_ms)
        safe_staleness = max(0.0, staleness_ms)
        with self._metrics_lock:
            self._plan_resolution_count += 1
            if resource_fallback_used:
                self._resource_fallback_count += 1

            self._staleness_count += 1
            self._staleness_ms_latest = safe_staleness
            self._staleness_ms_sum += safe_staleness
            if safe_staleness > self._staleness_ms_max:
                self._staleness_ms_max = safe_staleness

            self._latency_count += 1
            self._latency_ms_sum += safe_latency
            self._latency_ms_recent.append(safe_latency)
            if self._latency_ms_min is None or safe_latency < self._latency_ms_min:
                self._latency_ms_min = safe_latency
            if safe_latency > self._latency_ms_max:
                self._latency_ms_max = safe_latency

            for tier, count in locality_hits.items():
                if tier not in self._locality_hits:
                    self._locality_hits[tier] = 0
                self._locality_hits[tier] += count
                self._locality_total_hits += count


def _promote_node(nodes: list[ClusterNode], node_id: str) -> list[ClusterNode]:
    for idx, node in enumerate(nodes):
        if node.node_id == node_id:
            return [node, *nodes[:idx], *nodes[idx + 1 :]]
    return nodes


def _quantile(values: deque[float], q: float) -> float | None:
    if not values:
        return None
    if q <= 0.0:
        return min(values)
    if q >= 1.0:
        return max(values)
    samples = sorted(values)
    if len(samples) == 1:
        return float(samples[0])
    position = (len(samples) - 1) * float(q)
    lower = int(math.floor(position))
    upper = int(math.ceil(position))
    lower_value = float(samples[lower])
    upper_value = float(samples[upper])
    if lower == upper:
        return lower_value
    weight = position - float(lower)
    return lower_value + ((upper_value - lower_value) * weight)


def _dedupe_preserve_order(addresses: Iterable[str]) -> tuple[str, ...]:
    seen: set[str] = set()
    ordered: list[str] = []
    for address in addresses:
        if address in seen:
            continue
        seen.add(address)
        ordered.append(address)
    return tuple(ordered)


def _resource_match(
    summary: dict[str, object] | None,
    requirements: dict[str, int | float],
) -> bool:
    data = summary or {}
    for key, need in requirements.items():
        have = data.get(key)
        if not isinstance(have, (int, float)):
            return False
        if have < need:
            return False
    return True


def _owner_key(request: PlanResolutionRequest) -> str | None:
    if not request.lease_owner_kind or not request.lease_owner_id:
        return None
    return f"{request.lease_owner_kind}:{request.lease_owner_id}"


def _spillover_policy_enabled(request: PlanResolutionRequest) -> bool:
    return (
        request.spillover_max_remote_nodes is not None
        or request.spillover_ttl_ms is not None
        or not request.spillover_allow_remote
        or tuple(request.spillover_stage1_tiers) != ("local", "zone")
        or tuple(request.spillover_stage2_tiers) != ("region",)
        or request.spillover_claim_id is not None
    )


def _spillover_claim_key(request: PlanResolutionRequest) -> str:
    if request.spillover_claim_id:
        return request.spillover_claim_id
    owner_key = _owner_key(request)
    if owner_key:
        return f"{request.stage_id}:{owner_key}"
    return f"{request.stage_id}:{request.request_id}"


def _clamp_int(value: int, lower: int, upper: int) -> int:
    return max(lower, min(value, upper))


__all__ = [
    "DEFAULT_LOCALITY_ORDER",
    "PlanResolutionRequest",
    "PlanResolution",
    "ClusterPlanResolver",
]



from __future__ import annotations

import hashlib
import os
import socket
from dataclasses import dataclass
from typing import Callable, Iterable

from flownet.runtime.transport.packet_transport import EndpointRef, split_host_port


_LOCAL_HOST_ALIASES = frozenset({"localhost", "127.0.0.1", "::1", "0.0.0.0"})
_UDS_MODES = frozenset({"auto", "uds_prefer", "hetero_auto"})
_SHM_MODES = frozenset({"hetero_auto"})
_VALID_CAPS = frozenset({"tcp", "uds", "shm"})


@dataclass(frozen=True)
class TransportRoute:
    endpoint: EndpointRef
    backend_order: tuple[str, ...]


class TransportSelector:
    def __init__(
        self,
        *,
        local_host: str,
        local_port: int,
        transport_mode: str = "auto",
        local_ipc_dir: str = "/tmp/flownet",
        local_host_id: str | None = None,
        local_uds_path: str | None = None,
        local_shm_path: str | None = None,
        local_transport_caps: Iterable[str] | None = None,
        shm_threshold_bytes: int = 262144,
    ):
        self.local_host = local_host
        self.local_port = int(local_port)
        self.transport_mode = (transport_mode or "auto").strip().lower()
        self.local_ipc_dir = local_ipc_dir or "/tmp/flownet"
        self.local_host_id = local_host_id or self._build_default_host_id()
        self.local_transport_caps = self._normalize_transport_caps(
            local_transport_caps,
            transport_mode=self.transport_mode,
        )
        self.shm_threshold_bytes = max(1, int(shm_threshold_bytes))
        self.local_address = f"{self.local_host}:{self.local_port}"
        self.local_uds_path = local_uds_path or self.uds_path_for_local_port(self.local_port)
        self.local_shm_path = local_shm_path or self.shm_path_for_local_port(self.local_port)
        self._local_host_aliases = self._build_local_host_aliases(local_host)
        self._endpoint_lookup: Callable[[str], dict[str, object] | None] | None = None
        self._metrics = {
            "endpoint_lookup_hits": 0,
            "endpoint_lookup_misses": 0,
            "endpoint_lookup_errors": 0,
        }

    def set_endpoint_lookup(
        self,
        endpoint_lookup: Callable[[str], dict[str, object] | None] | None,
    ) -> None:
        self._endpoint_lookup = endpoint_lookup

    def resolve(self, address: str, payload_size_bytes: int | None = None) -> TransportRoute:
        host, port = split_host_port(address)
        local_host = self.is_local_host(host)
        endpoint_meta = self._resolve_endpoint_metadata(address)
        endpoint_host_id = endpoint_meta.get("host_id")
        if not isinstance(endpoint_host_id, str):
            endpoint_host_id = None
        endpoint_caps = self._normalize_transport_caps(
            endpoint_meta.get("transport_caps"),
            transport_mode=self.transport_mode,
            default_local=False,
            same_machine_hint=local_host,
        )
        endpoint_uds_path = endpoint_meta.get("uds_path")
        if not isinstance(endpoint_uds_path, str):
            endpoint_uds_path = self.uds_path_for_local_port(port) if local_host else self.uds_path_for_address(address)
        endpoint_shm_path = endpoint_meta.get("shm_path")
        if not isinstance(endpoint_shm_path, str):
            endpoint_shm_path = self.shm_path_for_local_port(port) if local_host else self.shm_path_for_address(address)
        same_machine = local_host or (
            isinstance(endpoint_host_id, str)
            and self.local_host_id is not None
            and endpoint_host_id == self.local_host_id
        )
        endpoint = EndpointRef(
            address=address,
            host=host,
            port=port,
            uds_path=endpoint_uds_path,
            host_id=endpoint_host_id,
            transport_caps=endpoint_caps,
            shm_path=endpoint_shm_path,
        )
        if self.transport_mode == "tcp":
            return TransportRoute(endpoint=endpoint, backend_order=("tcp",))
        uds_allowed = (
            same_machine
            and "uds" in self.local_transport_caps
            and "uds" in endpoint_caps
        )
        shm_allowed = (
            same_machine
            and "shm" in self.local_transport_caps
            and "shm" in endpoint_caps
            and payload_size_bytes is not None
            and payload_size_bytes >= self.shm_threshold_bytes
        )
        if (
            self.transport_mode in _SHM_MODES
            and shm_allowed
        ):
            backends: list[str] = ["shm"]
            if uds_allowed:
                backends.append("uds")
            backends.append("tcp")
            return TransportRoute(endpoint=endpoint, backend_order=tuple(backends))
        if self.transport_mode in _UDS_MODES and uds_allowed:
            return TransportRoute(endpoint=endpoint, backend_order=("uds", "tcp"))
        return TransportRoute(endpoint=endpoint, backend_order=("tcp",))

    def is_local_host(self, host: str) -> bool:
        return host.strip().lower() in self._local_host_aliases

    def uds_path_for_address(self, address: str) -> str:
        safe = address.replace("/", "_").replace(":", "_")
        digest = hashlib.sha1(address.encode("utf-8")).hexdigest()[:12]
        filename = f"flownet-{safe[:32]}-{digest}.sock"
        return os.path.join(self.local_ipc_dir, filename)

    def uds_path_for_local_port(self, port: int) -> str:
        # Canonical local IPC path should not depend on host spelling
        # (127.0.0.1 / localhost / 0.0.0.0 / hostname).
        return self.uds_path_for_address(f"local:{int(port)}")

    def shm_path_for_address(self, address: str) -> str:
        safe = address.replace("/", "_").replace(":", "_")
        digest = hashlib.sha1(address.encode("utf-8")).hexdigest()[:12]
        filename = f"flownet-shm-{safe[:32]}-{digest}.sock"
        return os.path.join(self.local_ipc_dir, filename)

    def shm_path_for_local_port(self, port: int) -> str:
        return self.shm_path_for_address(f"local:{int(port)}")

    def metrics_snapshot(self) -> dict[str, int]:
        return dict(self._metrics)

    def _resolve_endpoint_metadata(self, address: str) -> dict[str, object]:
        if self._endpoint_lookup is None:
            return {}
        try:
            payload = self._endpoint_lookup(address)
        except Exception:
            self._metrics["endpoint_lookup_errors"] += 1
            return {}
        if isinstance(payload, dict):
            self._metrics["endpoint_lookup_hits"] += 1
            return payload
        self._metrics["endpoint_lookup_misses"] += 1
        return {}

    @staticmethod
    def _build_default_host_id() -> str:
        try:
            return socket.gethostname().strip().lower()
        except Exception:
            return "unknown-host"

    @staticmethod
    def _normalize_transport_caps(
        transport_caps: Iterable[str] | None,
        *,
        transport_mode: str,
        default_local: bool = True,
        same_machine_hint: bool = False,
    ) -> tuple[str, ...]:
        normalized: list[str] = []
        if transport_caps is not None:
            source = (
                [item.strip() for item in transport_caps.split(",")]
                if isinstance(transport_caps, str)
                else transport_caps
            )
            for cap in source:
                value = str(cap).strip().lower()
                if value in _VALID_CAPS and value not in normalized:
                    normalized.append(value)
        if normalized:
            if "tcp" not in normalized:
                normalized.append("tcp")
            return tuple(normalized)
        if not default_local:
            if same_machine_hint and transport_mode != "tcp":
                if transport_mode == "hetero_auto":
                    return ("uds", "shm", "tcp")
                return ("uds", "tcp")
            return ("tcp",)
        if transport_mode == "tcp":
            return ("tcp",)
        if transport_mode == "hetero_auto":
            return ("uds", "shm", "tcp")
        return ("uds", "tcp")

    @staticmethod
    def _build_local_host_aliases(local_host: str) -> set[str]:
        aliases: set[str] = {local_host.strip().lower()}
        aliases.update(_LOCAL_HOST_ALIASES)
        try:
            infos = socket.getaddrinfo(socket.gethostname(), None)
            aliases.update(info[4][0].strip().lower() for info in infos if info[4])
        except Exception:
            pass
        try:
            infos = socket.getaddrinfo(local_host, None)
            aliases.update(info[4][0].strip().lower() for info in infos if info[4])
        except Exception:
            pass
        return aliases


__all__ = [
    "TransportRoute",
    "TransportSelector",
]



from __future__ import annotations

from dataclasses import dataclass, field
import time
from typing import Any


@dataclass
class ServiceEvent:
    service: str
    event_type: str
    payload: Any
    tags: dict[str, str] = field(default_factory=dict)
    source: str | None = None
    timestamp: float = field(default_factory=time.time)


__all__ = ["ServiceEvent"]



from flownet.runtime.flows.connectors import (
    INGRESS_MANUAL_PUSH,
    INGRESS_TOPIC,
    INGRESS_DATASET,
    EGRESS_DEFAULT,
    EGRESS_BUFFERED,
    EGRESS_TOPIC,
    EGRESS_RETURN_ONCE,
    EGRESS_EXTERNAL_SINK,
    IngressConnector,
    EgressConnector,
    resolve_ingress_connector,
    resolve_egress_connector,
)
from flownet.runtime.flows.run import (
    FlowRunRef,
    FlowProcess,
    RequestRef,
    FlowRunRegistry,
)
from flownet.runtime.flows.service import FlowRunService
from flownet.runtime.flows.program_catalog import FlowProgramRecord, FlowProgramCatalog
from flownet.runtime.flows.request_registry import (
    register_run_request,
    ack_run_request,
    reset_request_registry,
)
from flownet.runtime.flows.request_tracker import RunRequestTracker
from flownet.runtime.flows.stream_run_tracker import StreamRunTracker
from flownet.runtime.flows.topic_output import RunTopicEgressChannel
from flownet.runtime.flows.control_output import RunControlOutputCoordinator
from flownet.runtime.flows.replay_checkpoint import ReplayCheckpoint, parse_replay_checkpoint
from flownet.runtime.flows.ordered_event_channel import (
    EventEnvelope,
    TerminalEnvelope,
    GapNotice,
    OrderedEventChannel,
    OrderedEventSubscription,
)

__all__ = [
    "INGRESS_MANUAL_PUSH",
    "INGRESS_TOPIC",
    "INGRESS_DATASET",
    "EGRESS_DEFAULT",
    "EGRESS_BUFFERED",
    "EGRESS_TOPIC",
    "EGRESS_RETURN_ONCE",
    "EGRESS_EXTERNAL_SINK",
    "IngressConnector",
    "EgressConnector",
    "resolve_ingress_connector",
    "resolve_egress_connector",
    "FlowRunRef",
    "FlowProcess",
    "RequestRef",
    "FlowRunRegistry",
    "FlowRunService",
    "FlowProgramRecord",
    "FlowProgramCatalog",
    "register_run_request",
    "ack_run_request",
    "reset_request_registry",
    "RunRequestTracker",
    "StreamRunTracker",
    "RunTopicEgressChannel",
    "RunControlOutputCoordinator",
    "ReplayCheckpoint",
    "parse_replay_checkpoint",
    "EventEnvelope",
    "TerminalEnvelope",
    "GapNotice",
    "OrderedEventChannel",
    "OrderedEventSubscription",
]



from __future__ import annotations

import asyncio
from collections.abc import Coroutine
from typing import Any, Callable

from flownet.v1.runtime.comm import (
    OP_RPC_FLOW_PROGRAM_PULL,
    PLANE_RPC,
    V1CommHub,
)
from flownet.v1.runtime.flowengine.cursor_models import FlowProgramRef


def register_flow_program_pull_handler(
    *,
    comm_hub: V1CommHub,
    resolve_flow_program: Callable[[FlowProgramRef], Any | None],
) -> None:
    """Register owner-side rpc handler for flow-program pull requests."""

    def _handle_pull(envelope) -> dict[str, Any]:
        requested_ref = _decode_flow_program_ref_from_body(envelope.body)
        flow_program = resolve_flow_program(requested_ref)
        return {
            "mode": "flow_program_pull_result",
            "flow_program_ref": _encode_flow_program_ref(requested_ref),
            "found": flow_program is not None,
            "flow_program": flow_program,
        }

    comm_hub.protocol_router.register_handler(
        plane=PLANE_RPC,
        op=OP_RPC_FLOW_PROGRAM_PULL,
        handler=_handle_pull,
    )


def build_flow_program_pull_resolver(
    *,
    comm_hub: V1CommHub,
    resolve_owner_address: Callable[[FlowProgramRef], str],
    timeout: float | None = None,
    run_coroutine: Callable[[Coroutine[Any, Any, Any]], Any] | None = None,
) -> Callable[[FlowProgramRef], Any | None]:
    """Build sync FlowProgramCache pull callback from owner route + comm pull."""

    def _runner(coro: Coroutine[Any, Any, Any]) -> Any:
        if callable(run_coroutine):
            return run_coroutine(coro)
        try:
            asyncio.get_running_loop()
        except RuntimeError:
            return asyncio.run(coro)
        raise RuntimeError(
            "flow_program_pull_resolver_requires_run_coroutine_when_called_inside_running_event_loop",
        )

    async def _pull_async(flow_program_ref: FlowProgramRef) -> Any | None:
        owner_address = _normalize_non_empty(
            resolve_owner_address(flow_program_ref),
            field_name="owner_address",
        )
        return await pull_flow_program_via_comm(
            comm_hub=comm_hub,
            target_address=owner_address,
            flow_program_ref=flow_program_ref,
            timeout=timeout,
        )

    def _pull(flow_program_ref: FlowProgramRef) -> Any | None:
        normalized_ref = _normalize_program_ref(flow_program_ref)
        coro = _pull_async(normalized_ref)
        try:
            return _runner(coro)
        except Exception:
            try:
                coro.close()
            except Exception:
                pass
            raise

    return _pull


async def pull_flow_program_via_comm(
    *,
    comm_hub: V1CommHub,
    target_address: str,
    flow_program_ref: FlowProgramRef,
    timeout: float | None = None,
) -> Any | None:
    """Pull flow program from remote owner over rpc flow_program.pull."""

    normalized_target = _normalize_non_empty(target_address, field_name="target_address")
    normalized_ref = _normalize_program_ref(flow_program_ref)
    reply = await comm_hub.request(
        plane=PLANE_RPC,
        op=OP_RPC_FLOW_PROGRAM_PULL,
        target_address=normalized_target,
        timeout=timeout,
        request_ref_id=_normalize_request_ref(normalized_ref),
        body={
            "mode": "flow_program_pull",
            "flow_program_ref": _encode_flow_program_ref(normalized_ref),
        },
    )
    if not isinstance(reply.body, dict):
        raise TypeError("flow_program_pull_reply_body must be a dict.")
    body = dict(reply.body)

    if "error_type" in body and "message" in body and "mode" not in body:
        error_type = _normalize_non_empty(body.get("error_type"), field_name="error_type")
        message = _normalize_non_empty(body.get("message"), field_name="message")
        raise RuntimeError(f"flow_program_pull_failed:{error_type}:{message}")

    mode = _normalize_non_empty(body.get("mode"), field_name="mode")
    if mode != "flow_program_pull_result":
        raise ValueError(f"unsupported_flow_program_pull_reply_mode:{mode}")
    returned_ref = _decode_flow_program_ref_mapping(body.get("flow_program_ref"))
    if returned_ref != normalized_ref:
        raise ValueError(
            "flow_program_pull_ref_mismatch:"
            f" expected={_normalize_request_ref(normalized_ref)},"
            f" actual={_normalize_request_ref(returned_ref)}",
        )
    found = bool(body.get("found"))
    if not found:
        return None
    if "flow_program" not in body:
        raise ValueError("flow_program_pull_result_missing_flow_program")
    return body.get("flow_program")


def _decode_flow_program_ref_from_body(body: Any) -> FlowProgramRef:
    if not isinstance(body, dict):
        raise TypeError("flow_program_pull_body must be a dict.")
    mode = _normalize_non_empty(body.get("mode"), field_name="mode")
    if mode != "flow_program_pull":
        raise ValueError(f"unsupported_flow_program_pull_mode:{mode}")
    return _decode_flow_program_ref_mapping(body.get("flow_program_ref"))


def _decode_flow_program_ref_mapping(raw: Any) -> FlowProgramRef:
    if not isinstance(raw, dict):
        raise TypeError("flow_program_ref must be a dict.")
    return FlowProgramRef(
        program_uri=_normalize_non_empty(
            raw.get("program_uri"),
            field_name="flow_program_ref.program_uri",
        ),
        program_rev=_normalize_non_empty(
            raw.get("program_rev"),
            field_name="flow_program_ref.program_rev",
        ),
    )


def _encode_flow_program_ref(flow_program_ref: FlowProgramRef) -> dict[str, str]:
    normalized_ref = _normalize_program_ref(flow_program_ref)
    return {
        "program_uri": normalized_ref.program_uri,
        "program_rev": normalized_ref.program_rev,
    }


def _normalize_program_ref(flow_program_ref: FlowProgramRef) -> FlowProgramRef:
    if not isinstance(flow_program_ref, FlowProgramRef):
        raise TypeError("flow_program_ref must be FlowProgramRef.")
    return FlowProgramRef(
        program_uri=_normalize_non_empty(
            flow_program_ref.program_uri,
            field_name="flow_program_ref.program_uri",
        ),
        program_rev=_normalize_non_empty(
            flow_program_ref.program_rev,
            field_name="flow_program_ref.program_rev",
        ),
    )


def _normalize_request_ref(flow_program_ref: FlowProgramRef) -> str:
    normalized_ref = _normalize_program_ref(flow_program_ref)
    return f"{normalized_ref.program_uri}#{normalized_ref.program_rev}"


def _normalize_non_empty(value: Any, *, field_name: str) -> str:
    normalized = str(value or "").strip()
    if not normalized:
        raise ValueError(f"{field_name} must be non-empty.")
    return normalized


__all__ = [
    "register_flow_program_pull_handler",
    "build_flow_program_pull_resolver",
    "pull_flow_program_via_comm",
]



from __future__ import annotations

from typing import Any
from urllib.parse import urlparse

_FLOWNET_URI_SCHEME = "flownet://"
_TOPIC_PREFIX = "topic:"
_DEFAULT_CLUSTER = "local"
_DEFAULT_SCOPE = "default"
_TOPIC_KIND = "topic"


def _normalize_non_empty(value: str, *, field_name: str) -> str:
    normalized = str(value or "").strip()
    if not normalized:
        raise ValueError(f"{field_name} must be a non-empty string.")
    return normalized


def _normalize_non_negative_int(value: Any, *, field_name: str) -> int:
    try:
        normalized = int(value)
    except (TypeError, ValueError) as exc:
        raise ValueError(f"{field_name} must be an integer.") from exc
    if normalized < 0:
        raise ValueError(f"{field_name} must be non-negative.")
    return normalized


def _normalize_topic_uri(topic_uri: str) -> str:
    return _canonical_topic_key(_normalize_non_empty(topic_uri, field_name="topic_uri"))


def _canonical_topic_key(topic_id: str) -> str:
    raw_topic_id = _normalize_non_empty(topic_id, field_name="topic_id")
    if raw_topic_id.startswith(_FLOWNET_URI_SCHEME):
        return _normalize_flownet_topic_uri(raw_topic_id)
    if raw_topic_id.startswith(_TOPIC_PREFIX):
        topic_name = raw_topic_id[len(_TOPIC_PREFIX) :].strip()
        if not topic_name:
            raise ValueError("topic shorthand must include a non-empty name.")
        return _build_topic_uri(topic_name)
    if ":" in raw_topic_id:
        return raw_topic_id
    return _build_topic_uri(raw_topic_id)


def _build_topic_uri(topic_name: str) -> str:
    return (
        f"{_FLOWNET_URI_SCHEME}"
        f"{_DEFAULT_CLUSTER}/{_DEFAULT_SCOPE}/{_TOPIC_KIND}/{topic_name}"
    )


def _normalize_flownet_topic_uri(topic_uri: str) -> str:
    parsed = urlparse(topic_uri)
    if parsed.scheme != "flownet":
        raise ValueError(f"unsupported topic URI scheme: {parsed.scheme!r}")
    if not parsed.netloc:
        raise ValueError("topic URI must include a cluster segment.")
    path_segments = [segment for segment in parsed.path.split("/") if segment]
    if len(path_segments) < 3:
        raise ValueError(
            "topic URI must follow flownet://<cluster>/<scope>/topic/<name>.",
        )
    scope, kind, *name_segments = path_segments
    if kind != _TOPIC_KIND:
        return topic_uri
    topic_name = "/".join(name_segments).strip("/")
    if not topic_name:
        raise ValueError("topic URI must include a non-empty topic name.")
    return f"{_FLOWNET_URI_SCHEME}{parsed.netloc}/{scope}/{_TOPIC_KIND}/{topic_name}"


__all__ = [
    "_normalize_non_empty",
    "_normalize_non_negative_int",
    "_normalize_topic_uri",
]



from __future__ import annotations

import contextvars
from dataclasses import dataclass
import inspect
import threading
import uuid
from typing import Any, Callable

from flownet.core.request import FlownetPacket
from flownet.core.service_event import ServiceEvent
from flownet.core.types import PayloadData
from flownet.compiler.flowtask import FlowTask
from flownet.runtime.contexts.actor import _system_calls_scope, _user_execution_scope
from flownet.runtime.loops import submit_runtime_coro
from flownet.runtime.runner_lifecycle import start_runner_lifecycle


_service_context: contextvars.ContextVar["ServiceContext" | None] = contextvars.ContextVar(
    "flownet_service_context",
    default=None,
)


class ServiceContext:
    def __init__(
        self,
        name: str,
        registry: "ServiceRegistry",
        call_manager: Any,
        local_address_fn: Callable[[], str],
        stop_event: threading.Event,
    ):
        self._name = name
        self._registry = registry
        self._call_manager = call_manager
        self._local_address_fn = local_address_fn
        self._stop_event = stop_event

    @property
    def name(self) -> str:
        return self._name

    def emit(
        self,
        event_type: str,
        payload: Any = None,
        tags: dict[str, str] | None = None,
    ) -> None:
        with _system_calls_scope(True):
            self._registry.publish(self._name, event_type, payload, tags)

    def local_address(self) -> str:
        return self._local_address_fn()

    def should_stop(self) -> bool:
        return self._stop_event.is_set()

    @property
    def stop_event(self) -> threading.Event:
        return self._stop_event

    def submit(self, coro: Any, *, loop: str = "user"):
        return self._registry.submit(coro, loop=loop)


def service_context() -> ServiceContext:
    context = _service_context.get()
    if context is None:
        raise RuntimeError("Service context not available. Use within a service runner.")
    return context


def emit(
    event_type: str,
    payload: Any = None,
    tags: dict[str, str] | None = None,
) -> None:
    service_context().emit(event_type, payload, tags)


@dataclass
class ServiceHandle:
    name: str
    service: Any
    status: str = "running"
    restart_policy: str = "none"
    last_error: Exception | None = None
    runner: "ServiceRunner" | None = None


@dataclass(frozen=True)
class FlowSubscription:
    subscription_id: str
    service_name: str
    flow_task: FlowTask
    execution_address: str
    tags: dict[str, str] | None = None


class ServiceRunner:
    def __init__(
        self,
        name: str,
        stop_event: threading.Event,
        thread: threading.Thread | None = None,
        future: Any | None = None,
        close_fn: Callable[[], None] | None = None,
    ):
        self._name = name
        self._stop_event = stop_event
        self._thread = thread
        self._future = future
        self._close_fn = close_fn

    def stop(self) -> None:
        self._stop_event.set()
        if self._close_fn is not None:
            try:
                self._close_fn()
            except Exception as exc:  # pragma: no cover - best effort cleanup
                print(f"[ServiceRunner] Failed to close {self._name}: {exc}")
        if self._future is not None:
            try:
                self._future.cancel()
            except Exception as exc:  # pragma: no cover - best effort
                print(f"[ServiceRunner] Failed to cancel {self._name}: {exc}")
        if self._thread is not None and self._thread.is_alive():
            self._thread.join(timeout=1.0)


def _emit_item(context: ServiceContext, item: Any) -> None:
    if item is None:
        return
    if isinstance(item, ServiceEvent):
        context.emit(item.event_type, item.payload, dict(item.tags))
        return
    if isinstance(item, tuple):
        if len(item) == 2:
            event_type, payload = item
            tags = None
        elif len(item) == 3:
            event_type, payload, tags = item
        else:
            raise ValueError(f"Invalid service event tuple: {item}")
        context.emit(event_type, payload, tags)
        return
    context.emit("event", item, None)


def _call_service(
    service_spec: Any,
    *args,
    **kwargs,
) -> tuple[Any, Any]:
    if inspect.isclass(service_spec):
        service_obj = service_spec(*args, **kwargs)
        result = service_obj()
        return service_obj, result
    if inspect.isgenerator(service_spec) or inspect.isasyncgen(service_spec):
        return service_spec, service_spec
    if not callable(service_spec):
        raise TypeError(f"Service spec is not callable: {type(service_spec)}")
    result = service_spec(*args, **kwargs)
    return service_spec, result


class ServiceRegistry:
    def __init__(
        self,
        call_manager: Any,
        local_address_fn: Callable[[], str],
        submit_coro: Callable[[Any], Any] = submit_runtime_coro,
        submit_user_coro: Callable[[Any], Any] | None = None,
        topic_registry: Any | None = None,
    ):
        self._call_manager = call_manager
        self._local_address_fn = local_address_fn
        self._submit_coro = submit_coro
        self._submit_user_coro = submit_user_coro or submit_coro
        self._topic_registry = topic_registry
        self._services: dict[str, ServiceHandle] = {}
        self._flow_subscriptions: dict[str, dict[str, FlowSubscription]] = {}
        self._subscriptions_by_id: dict[str, FlowSubscription] = {}
        self._lock = threading.Lock()

    def start_service(
        self,
        name: str,
        service_spec: Any,
        *args,
        restart_policy: str = "none",
        **kwargs,
    ) -> ServiceHandle:
        with self._lock:
            if name in self._services:
                raise ValueError(f"Service already running: {name}")

        stop_event = threading.Event()
        context = ServiceContext(
            name=name,
            registry=self,
            call_manager=self._call_manager,
            local_address_fn=self._local_address_fn,
            stop_event=stop_event,
        )

        service_obj, result = _call_service(service_spec, *args, **kwargs)

        handle = ServiceHandle(
            name=name,
            service=service_obj,
            restart_policy=restart_policy,
        )
        with self._lock:
            self._services[name] = handle

        try:
            runner = self._start_runner(name, context, stop_event, result, handle)
            handle.runner = runner
        except Exception:
            with self._lock:
                self._services.pop(name, None)
            raise
        return handle

    def _start_runner(
        self,
        name: str,
        context: ServiceContext,
        stop_event: threading.Event,
        result: Any,
        handle: ServiceHandle,
    ) -> ServiceRunner | None:
        def _on_stopped(_runner_kind: str) -> None:
            self._update_status(name, "stopped", None)

        def _on_cancelled() -> None:
            self._update_status(name, "stopped", None)

        def _on_error(exc: Exception) -> None:
            self._update_status(name, "error", exc)
            print(f"[ServiceRegistry] Service {name} failed: {exc}")

        started = start_runner_lifecycle(
            name=name,
            result=result,
            stop_event=stop_event,
            submit_user_coro=self._submit_user_coro,
            set_context=lambda: _service_context.set(context),
            reset_context=_service_context.reset,
            execution_scope_factory=lambda: _user_execution_scope(None),
            emit_item=lambda item: _emit_item(context, item),
            on_stopped=_on_stopped,
            on_error=_on_error,
            on_cancelled=_on_cancelled,
            thread_name=f"svc-{name}",
        )
        if started is None:
            return None
        return ServiceRunner(
            name=name,
            stop_event=stop_event,
            thread=started.thread,
            future=started.future,
            close_fn=started.close_fn,
        )

    def submit(self, coro: Any, *, loop: str = "runtime"):
        if loop == "user" and self._submit_user_coro is not None:
            return self._submit_user_coro(coro)
        return self._submit_coro(coro)

    def _update_status(self, name: str, status: str, error: Exception | None) -> None:
        with self._lock:
            handle = self._services.get(name)
            if handle is None:
                return
            handle.status = status
            handle.last_error = error

    def stop_service(self, name: str) -> bool:
        with self._lock:
            handle = self._services.get(name)
        if handle is None:
            return False
        if handle.runner is not None:
            handle.runner.stop()
        stop_method = getattr(handle.service, "stop", None)
        if callable(stop_method):
            stop_result = stop_method()
            if inspect.isawaitable(stop_result):
                self._submit_user_coro(stop_result)
        with self._lock:
            self._services.pop(name, None)
        return True

    def list_services(self) -> list[dict[str, Any]]:
        with self._lock:
            handles = list(self._services.values())
        return [
            {
                "name": handle.name,
                "status": handle.status,
                "restart_policy": handle.restart_policy,
                "last_error": str(handle.last_error) if handle.last_error else None,
            }
            for handle in handles
        ]

    def get_service(self, name: str) -> ServiceHandle | None:
        with self._lock:
            return self._services.get(name)

    def subscribe_flow(
        self,
        service_name: str,
        flow_task: FlowTask,
        execution_address: str,
        tags: dict[str, str] | None = None,
    ) -> str:
        if self._topic_registry is not None:
            return self._topic_registry.attach_task(
                service_topic_id(service_name),
                flow_task,
                execution_address,
                tags,
            )
        subscription_id = uuid.uuid4().hex
        subscription = FlowSubscription(
            subscription_id=subscription_id,
            service_name=service_name,
            flow_task=flow_task,
            execution_address=execution_address,
            tags=tags,
        )
        with self._lock:
            self._flow_subscriptions.setdefault(service_name, {})[subscription_id] = subscription
            self._subscriptions_by_id[subscription_id] = subscription
        return subscription_id

    def unsubscribe_flow(self, subscription_id: str) -> bool:
        if self._topic_registry is not None:
            return self._topic_registry.detach_task(subscription_id)
        with self._lock:
            subscription = self._subscriptions_by_id.pop(subscription_id, None)
            if subscription is None:
                return False
            subscriptions = self._flow_subscriptions.get(subscription.service_name)
            if subscriptions is not None:
                subscriptions.pop(subscription_id, None)
                if not subscriptions:
                    self._flow_subscriptions.pop(subscription.service_name, None)
        return True

    def publish(
        self,
        service_name: str,
        event_type: str,
        payload: Any = None,
        tags: dict[str, str] | None = None,
    ) -> None:
        if self._topic_registry is not None:
            service_topic = service_topic_id(service_name)
            self._topic_registry.publish(
                service_topic,
                event_type,
                payload,
                tags,
                producer_id=service_topic,
                allow_reserved_namespace=True,
            )
            return
        if tags is None:
            tags = {}
        event = ServiceEvent(
            service=service_name,
            event_type=event_type,
            payload=payload,
            tags=tags,
            source=self._local_address_fn(),
        )
        self._dispatch_flow(event)

    def _dispatch_flow(self, event: ServiceEvent) -> None:
        with self._lock:
            subscriptions = list(self._flow_subscriptions.get(event.service, {}).values())
        if not subscriptions:
            return

        async def _send():
            for subscription in subscriptions:
                payload = PayloadData(args=[event], kwargs={})
                request = subscription.flow_task.make_request(
                    payload,
                    origin_address=self._local_address_fn(),
                )
                if subscription.execution_address == self._local_address_fn():
                    try:
                        handler = getattr(self._call_manager, "runtime_handler", None)
                        if handler is not None:
                            await handler(
                                FlownetPacket(
                                    type="flowrequest",
                                    uuid=uuid.uuid4().hex,
                                    reply=False,
                                    packet_content=request,
                                )
                            )
                            continue
                    except Exception as exc:  # pragma: no cover - best effort
                        print("[ServiceRegistry] Local flow dispatch failed:", exc)
                packet = FlownetPacket(
                    type="flowrequest",
                    uuid=uuid.uuid4().hex,
                    reply=False,
                    packet_content=request,
                )
                try:
                    await self._call_manager.send_packet(subscription.execution_address, packet)
                except Exception as exc:  # pragma: no cover - best effort
                    print(
                        "[ServiceRegistry] Failed to send flowrequest to",
                        subscription.execution_address,
                        exc,
                    )

        self._submit_coro(_send())

    def handle_service_event(self, event: ServiceEvent) -> None:
        if self._topic_registry is not None:
            service_topic = service_topic_id(event.service)
            self._topic_registry.publish(
                service_topic,
                event.event_type,
                event.payload,
                dict(event.tags),
                producer_id=service_topic,
                allow_reserved_namespace=True,
            )
            return
        self._dispatch_flow(event)


def service_topic_id(service_name: str) -> str:
    return f"service:{service_name}"


__all__ = [
    "ServiceContext",
    "ServiceHandle",
    "ServiceRegistry",
    "ServiceRunner",
    "FlowSubscription",
    "service_topic_id",
    "service_context",
    "emit",
]



from __future__ import annotations

from collections.abc import Iterable
from typing import Any

from flownet.core.types import PayloadData


class ResultNormalizer:
    """Normalize operator results into PayloadData lists."""

    def normalize(self, result: Any, expand_iterable: bool) -> list[PayloadData]:
        if result is None:
            return []

        if isinstance(result, PayloadData):
            return [result]

        if isinstance(result, list) and all(
            isinstance(item, PayloadData) for item in result
        ):
            return result

        if isinstance(result, (str, bytes)):
            return [PayloadData(args=[result], kwargs={})]

        if expand_iterable and isinstance(result, Iterable):
            payloads: list[PayloadData] = []
            for item in result:
                if isinstance(item, PayloadData):
                    payloads.append(item)
                else:
                    payloads.append(PayloadData(args=[item], kwargs={}))
            return payloads

        return [PayloadData(args=[result], kwargs={})]


__all__ = ["ResultNormalizer"]



from __future__ import annotations

from contextlib import contextmanager

from flownet.utils.context_vars import ContextSlot


_flow_build_ctx = ContextSlot[bool]("flownet_flow_building", default=False)


def in_flow_build() -> bool:
    return _flow_build_ctx.get()


@contextmanager
def flow_build_context():
    with _flow_build_ctx.use(True):
        yield



from flownet.platform.user.platform_client_actor import PlatformClientActor
from flownet.platform.user.startup import (
    BootConfig,
    boot_user_runtime,
    ensure_node_runtime,
    auto_boot_user_runtime,
    DEFAULT_NODE_ADDRESS,
)

__all__ = [
    "PlatformClientActor",
    "BootConfig",
    "boot_user_runtime",
    "ensure_node_runtime",
    "auto_boot_user_runtime",
    "DEFAULT_NODE_ADDRESS",
]



from __future__ import annotations

from typing import Any, Protocol


class StateBackendError(RuntimeError):
    """Base exception for state backend failures."""


class StateCodecCompatibilityError(StateBackendError):
    """Raised when stored state is incompatible with current runtime codec version."""


class StateBackend(Protocol):
    def get(self, namespace: str, key: str) -> Any | None:
        ...

    def put(
        self,
        namespace: str,
        key: str,
        value: Any,
        ttl_seconds: float | None = None,
    ) -> None:
        ...

    def delete(self, namespace: str, key: str) -> None:
        ...

    def scan(
        self,
        namespace: str,
        *,
        prefix: str | None = None,
        limit: int | None = None,
    ) -> list[tuple[str, Any]]:
        ...

    def close(self) -> None:
        ...


__all__ = [
    "StateBackend",
    "StateBackendError",
    "StateCodecCompatibilityError",
]



from __future__ import annotations

import asyncio
import os
import uuid
from typing import Any, Awaitable, Callable

from flownet.api.actor_method_ref import ActorMethodRef
from flownet.core.request import ActorCall, FlownetPacket
from flownet.core.types import PayloadData
from flownet.runtime.transport.call_manager import RemoteCallManager
from flownet.runtime.actors.execution import ActorCallExecutor


def _rpc_debug_logs_enabled() -> bool:
    value = (os.environ.get("FLOWNET_RPC_DEBUG_LOGS") or "").strip().lower()
    return value in {"1", "true", "yes", "on"}


class ActorDispatcher:
    def __init__(
        self,
        call_manager: RemoteCallManager,
        local_address_fn: Callable[[], str],
        resolve_actor_fn: Callable[[str], Any],
        user_loop: asyncio.AbstractEventLoop | None,
        container_handler: Any,
        call_remote_via_v1: Callable[[ActorMethodRef, tuple[Any, ...], dict[str, Any]], Awaitable[Any]] | None = None,
        v1_fail_open_legacy: bool = True,
    ):
        self._call_manager = call_manager
        self._local_address_fn = local_address_fn
        self._resolve_actor_fn = resolve_actor_fn
        self._user_loop = user_loop
        self._container_handler = container_handler
        self._call_remote_via_v1 = call_remote_via_v1
        self._v1_fail_open_legacy = bool(v1_fail_open_legacy)
        self._executor = ActorCallExecutor(
            resolve_actor_fn=resolve_actor_fn,
            user_loop=user_loop,
            container_handler=container_handler,
        )

    async def handle_actor_call(self, rmf: ActorMethodRef, *args, **kwargs) -> Any:
        if _rpc_debug_logs_enabled():
            print("[FlownetRuntime] Handling actor call:", rmf, args, kwargs)
        local_address = self._local_address_fn()
        if rmf.address != local_address:
            if _rpc_debug_logs_enabled():
                print("[FlownetRuntime] Routing actor call to remote address:", rmf.address)
            if self._call_remote_via_v1 is not None:
                try:
                    return await self._call_remote_via_v1(
                        rmf,
                        tuple(args),
                        dict(kwargs),
                    )
                except Exception:
                    if not self._v1_fail_open_legacy:
                        raise
            return await self._call_manager.send_packet(
                rmf.address,
                FlownetPacket(
                    type="actorcall",
                    uuid=uuid.uuid4().hex,
                    reply=True,
                    packet_content=ActorCall(
                        source=local_address,
                        destination=rmf,
                        payload=PayloadData(args=list(args), kwargs=kwargs),
                    ),
                ),
            )

        return await self._executor.execute(rmf, *args, **kwargs)



from __future__ import annotations

from dataclasses import dataclass
from typing import Any


MSG_DELTA_PUSH = "cv.delta.push"
MSG_DELTA_PULL = "cv.delta.pull"
MSG_DELTA_ACK = "cv.delta.ack"
MSG_SNAPSHOT = "cv.snapshot"

SUPPORTED_SCHEMA = "v1"
SUPPORTED_MSG_TYPES = frozenset(
    {
        MSG_DELTA_PUSH,
        MSG_DELTA_PULL,
        MSG_DELTA_ACK,
        MSG_SNAPSHOT,
    }
)


@dataclass(frozen=True)
class ClusterViewDelta:
    membership: dict[str, dict[str, Any]]
    tombstones: dict[str, int] | None = None
    leases_summary: dict[str, Any] | None = None

    def as_dict(self) -> dict[str, Any]:
        payload: dict[str, Any] = {
            "membership": dict(self.membership),
        }
        if self.tombstones is not None:
            payload["tombstones"] = dict(self.tombstones)
        if self.leases_summary is not None:
            payload["leases_summary"] = dict(self.leases_summary)
        return payload


@dataclass(frozen=True)
class ClusterViewGossipEnvelope:
    msg_type: str
    schema_version: str
    from_node: str
    to_node: str
    view_epoch: int
    delta_seq: int
    sent_at_ms: int
    delta: ClusterViewDelta | None = None

    def as_dict(self) -> dict[str, Any]:
        payload: dict[str, Any] = {
            "msg_type": self.msg_type,
            "schema_version": self.schema_version,
            "from_node": self.from_node,
            "to_node": self.to_node,
            "view_epoch": self.view_epoch,
            "delta_seq": self.delta_seq,
            "sent_at_ms": self.sent_at_ms,
        }
        if self.delta is not None:
            payload["delta"] = self.delta.as_dict()
        return payload


def parse_envelope(
    payload: ClusterViewGossipEnvelope | dict[str, Any],
    *,
    msg_type: str | None = None,
) -> ClusterViewGossipEnvelope:
    if isinstance(payload, ClusterViewGossipEnvelope):
        return payload
    if not isinstance(payload, dict):
        raise TypeError("Cluster gossip payload must be dict or ClusterViewGossipEnvelope.")

    envelope_type = payload.get("msg_type", msg_type)
    if not isinstance(envelope_type, str):
        raise ValueError("cluster gossip payload missing msg_type.")

    raw_delta = payload.get("delta")
    delta: ClusterViewDelta | None = None
    if raw_delta is not None:
        if not isinstance(raw_delta, dict):
            raise TypeError("cluster gossip delta must be a dict.")
        membership = raw_delta.get("membership")
        if not isinstance(membership, dict):
            raise ValueError("cluster gossip delta missing membership map.")
        tombstones = raw_delta.get("tombstones")
        leases_summary = raw_delta.get("leases_summary")
        delta = ClusterViewDelta(
            membership=membership,
            tombstones=tombstones if isinstance(tombstones, dict) else None,
            leases_summary=leases_summary if isinstance(leases_summary, dict) else None,
        )

    return ClusterViewGossipEnvelope(
        msg_type=envelope_type,
        schema_version=str(payload.get("schema_version", "")),
        from_node=str(payload.get("from_node", "")),
        to_node=str(payload.get("to_node", "")),
        view_epoch=int(payload.get("view_epoch", -1)),
        delta_seq=int(payload.get("delta_seq", -1)),
        sent_at_ms=int(payload.get("sent_at_ms", 0)),
        delta=delta,
    )


def make_delta_push(
    *,
    from_node: str,
    to_node: str,
    view_epoch: int,
    delta_seq: int,
    sent_at_ms: int,
    membership: dict[str, dict[str, Any]],
    tombstones: dict[str, int] | None = None,
    leases_summary: dict[str, Any] | None = None,
    schema_version: str = SUPPORTED_SCHEMA,
) -> ClusterViewGossipEnvelope:
    return ClusterViewGossipEnvelope(
        msg_type=MSG_DELTA_PUSH,
        schema_version=schema_version,
        from_node=from_node,
        to_node=to_node,
        view_epoch=view_epoch,
        delta_seq=delta_seq,
        sent_at_ms=sent_at_ms,
        delta=ClusterViewDelta(
            membership=membership,
            tombstones=tombstones,
            leases_summary=leases_summary,
        ),
    )


def make_delta_pull(
    *,
    from_node: str,
    to_node: str,
    view_epoch: int,
    delta_seq: int,
    sent_at_ms: int,
    schema_version: str = SUPPORTED_SCHEMA,
) -> ClusterViewGossipEnvelope:
    return ClusterViewGossipEnvelope(
        msg_type=MSG_DELTA_PULL,
        schema_version=schema_version,
        from_node=from_node,
        to_node=to_node,
        view_epoch=view_epoch,
        delta_seq=delta_seq,
        sent_at_ms=sent_at_ms,
        delta=None,
    )


def make_delta_ack(
    *,
    from_node: str,
    to_node: str,
    view_epoch: int,
    delta_seq: int,
    sent_at_ms: int,
    schema_version: str = SUPPORTED_SCHEMA,
) -> ClusterViewGossipEnvelope:
    return ClusterViewGossipEnvelope(
        msg_type=MSG_DELTA_ACK,
        schema_version=schema_version,
        from_node=from_node,
        to_node=to_node,
        view_epoch=view_epoch,
        delta_seq=delta_seq,
        sent_at_ms=sent_at_ms,
        delta=None,
    )


def make_snapshot(
    *,
    from_node: str,
    to_node: str,
    view_epoch: int,
    delta_seq: int,
    sent_at_ms: int,
    membership: dict[str, dict[str, Any]],
    tombstones: dict[str, int] | None = None,
    leases_summary: dict[str, Any] | None = None,
    schema_version: str = SUPPORTED_SCHEMA,
) -> ClusterViewGossipEnvelope:
    return ClusterViewGossipEnvelope(
        msg_type=MSG_SNAPSHOT,
        schema_version=schema_version,
        from_node=from_node,
        to_node=to_node,
        view_epoch=view_epoch,
        delta_seq=delta_seq,
        sent_at_ms=sent_at_ms,
        delta=ClusterViewDelta(
            membership=membership,
            tombstones=tombstones,
            leases_summary=leases_summary,
        ),
    )


def validate_envelope(envelope: ClusterViewGossipEnvelope) -> None:
    if envelope.msg_type not in SUPPORTED_MSG_TYPES:
        raise ValueError(f"Unsupported gossip msg_type: {envelope.msg_type}")
    if envelope.schema_version != SUPPORTED_SCHEMA:
        raise ValueError(
            f"Unsupported gossip schema_version: {envelope.schema_version}"
        )
    if envelope.view_epoch < 0:
        raise ValueError("view_epoch must be >= 0")
    if envelope.delta_seq < 0:
        raise ValueError("delta_seq must be >= 0")
    if envelope.sent_at_ms <= 0:
        raise ValueError("sent_at_ms must be > 0")
    if envelope.msg_type in {MSG_DELTA_PUSH, MSG_SNAPSHOT} and envelope.delta is None:
        raise ValueError(f"{envelope.msg_type} requires delta payload")


__all__ = [
    "MSG_DELTA_PUSH",
    "MSG_DELTA_PULL",
    "MSG_DELTA_ACK",
    "MSG_SNAPSHOT",
    "SUPPORTED_SCHEMA",
    "SUPPORTED_MSG_TYPES",
    "ClusterViewDelta",
    "ClusterViewGossipEnvelope",
    "parse_envelope",
    "make_delta_push",
    "make_delta_pull",
    "make_delta_ack",
    "make_snapshot",
    "validate_envelope",
]



from __future__ import annotations

from collections.abc import Mapping, Sequence
from typing import Any

from flownet.v1.runtime.flowengine.cursor_models import (
    EventCursor,
    FlowProgramRef,
    FlowStackFrame,
    TopicRef,
)

_CURSOR_KEYS = frozenset(
    {
        "event_group_id",
        "event_chain_id",
        "flow_program_ref",
        "pc_node_id",
        "flow_stack",
        "payload",
        "tags",
        "seq",
        "origin_topic_ref",
        "convergence_topic_ref",
        "convergence_topic_epoch",
        "meta",
    }
)
_FLOW_PROGRAM_REF_KEYS = frozenset({"program_uri", "program_rev"})
_TOPIC_REF_KEYS = frozenset({"topic_uri", "coordinator_address"})
_FRAME_KEYS = frozenset(
    {
        "resume_flow_program_ref",
        "resume_pc_node_ids",
        "caller_event_chain_id",
        "callsite_pc_node_id",
        "frame_meta",
    }
)


class EventCursorDecodeError(ValueError):
    """Raised when incoming cursor payload violates canonical schema."""


def decode_event_cursor(raw: Mapping[str, Any]) -> EventCursor:
    mapping = _require_mapping(raw, field_name="event_cursor")
    _assert_exact_keys(mapping, _CURSOR_KEYS, field_name="event_cursor")

    return EventCursor(
        event_group_id=_require_non_empty_str(mapping["event_group_id"], field_name="event_group_id"),
        event_chain_id=_require_non_empty_str(mapping["event_chain_id"], field_name="event_chain_id"),
        flow_program_ref=_decode_flow_program_ref(mapping["flow_program_ref"]),
        pc_node_id=_require_non_empty_str(mapping["pc_node_id"], field_name="pc_node_id"),
        flow_stack=_decode_flow_stack(mapping["flow_stack"]),
        payload=mapping["payload"],
        tags=_decode_string_dict(mapping["tags"], field_name="tags"),
        seq=_decode_seq(mapping["seq"]),
        origin_topic_ref=_decode_topic_ref(mapping["origin_topic_ref"], field_name="origin_topic_ref"),
        convergence_topic_ref=_decode_topic_ref(
            mapping["convergence_topic_ref"],
            field_name="convergence_topic_ref",
        ),
        convergence_topic_epoch=_require_non_negative_int(
            mapping["convergence_topic_epoch"],
            field_name="convergence_topic_epoch",
        ),
        meta=_decode_meta(mapping["meta"], field_name="meta"),
    )


def encode_event_cursor(cursor: EventCursor) -> dict[str, Any]:
    return {
        "event_group_id": cursor.event_group_id,
        "event_chain_id": cursor.event_chain_id,
        "flow_program_ref": {
            "program_uri": cursor.flow_program_ref.program_uri,
            "program_rev": cursor.flow_program_ref.program_rev,
        },
        "pc_node_id": cursor.pc_node_id,
        "flow_stack": [
            {
                "resume_flow_program_ref": {
                    "program_uri": frame.resume_flow_program_ref.program_uri,
                    "program_rev": frame.resume_flow_program_ref.program_rev,
                },
                "resume_pc_node_ids": list(frame.resume_pc_node_ids),
                "caller_event_chain_id": frame.caller_event_chain_id,
                "callsite_pc_node_id": frame.callsite_pc_node_id,
                "frame_meta": dict(frame.frame_meta),
            }
            for frame in cursor.flow_stack
        ],
        "payload": cursor.payload,
        "tags": dict(cursor.tags),
        "seq": cursor.seq,
        "origin_topic_ref": {
            "topic_uri": cursor.origin_topic_ref.topic_uri,
            "coordinator_address": cursor.origin_topic_ref.coordinator_address,
        },
        "convergence_topic_ref": {
            "topic_uri": cursor.convergence_topic_ref.topic_uri,
            "coordinator_address": cursor.convergence_topic_ref.coordinator_address,
        },
        "convergence_topic_epoch": cursor.convergence_topic_epoch,
        "meta": dict(cursor.meta),
    }


def _decode_flow_stack(raw: Any) -> tuple[FlowStackFrame, ...]:
    if not isinstance(raw, Sequence) or isinstance(raw, (str, bytes, bytearray)):
        raise EventCursorDecodeError("flow_stack must be a sequence.")
    frames: list[FlowStackFrame] = []
    for index, item in enumerate(raw):
        frame_mapping = _require_mapping(item, field_name=f"flow_stack[{index}]")
        _assert_exact_keys(frame_mapping, _FRAME_KEYS, field_name=f"flow_stack[{index}]")
        resume_nodes_raw = frame_mapping["resume_pc_node_ids"]
        if not isinstance(resume_nodes_raw, Sequence) or isinstance(
            resume_nodes_raw,
            (str, bytes, bytearray),
        ):
            raise EventCursorDecodeError(f"flow_stack[{index}].resume_pc_node_ids must be a sequence.")
        resume_pc_node_ids = tuple(
            _require_non_empty_str(
                node_id,
                field_name=f"flow_stack[{index}].resume_pc_node_ids[{node_index}]",
            )
            for node_index, node_id in enumerate(resume_nodes_raw)
        )
        if not resume_pc_node_ids:
            raise EventCursorDecodeError(f"flow_stack[{index}].resume_pc_node_ids must not be empty.")

        frames.append(
            FlowStackFrame(
                resume_flow_program_ref=_decode_flow_program_ref(
                    frame_mapping["resume_flow_program_ref"],
                    field_name=f"flow_stack[{index}].resume_flow_program_ref",
                ),
                resume_pc_node_ids=resume_pc_node_ids,
                caller_event_chain_id=_require_non_empty_str(
                    frame_mapping["caller_event_chain_id"],
                    field_name=f"flow_stack[{index}].caller_event_chain_id",
                ),
                callsite_pc_node_id=_require_non_empty_str(
                    frame_mapping["callsite_pc_node_id"],
                    field_name=f"flow_stack[{index}].callsite_pc_node_id",
                ),
                frame_meta=_decode_meta(
                    frame_mapping["frame_meta"],
                    field_name=f"flow_stack[{index}].frame_meta",
                ),
            )
        )
    return tuple(frames)


def _decode_flow_program_ref(
    raw: Any,
    *,
    field_name: str = "flow_program_ref",
) -> FlowProgramRef:
    mapping = _require_mapping(raw, field_name=field_name)
    _assert_exact_keys(mapping, _FLOW_PROGRAM_REF_KEYS, field_name=field_name)
    return FlowProgramRef(
        program_uri=_require_non_empty_str(mapping["program_uri"], field_name=f"{field_name}.program_uri"),
        program_rev=_require_non_empty_str(mapping["program_rev"], field_name=f"{field_name}.program_rev"),
    )


def _decode_topic_ref(raw: Any, *, field_name: str) -> TopicRef:
    mapping = _require_mapping(raw, field_name=field_name)
    _assert_exact_keys(mapping, _TOPIC_REF_KEYS, field_name=field_name)
    return TopicRef(
        topic_uri=_require_non_empty_str(mapping["topic_uri"], field_name=f"{field_name}.topic_uri"),
        coordinator_address=_require_non_empty_str(
            mapping["coordinator_address"],
            field_name=f"{field_name}.coordinator_address",
        ),
    )


def _decode_seq(raw: Any) -> int | None:
    if raw is None:
        return None
    return _require_non_negative_int(raw, field_name="seq")


def _decode_meta(raw: Any, *, field_name: str) -> dict[str, Any]:
    if not isinstance(raw, Mapping):
        raise EventCursorDecodeError(f"{field_name} must be a mapping.")
    resolved: dict[str, Any] = {}
    for key, value in raw.items():
        if not isinstance(key, str) or not key.strip():
            raise EventCursorDecodeError(f"{field_name} keys must be non-empty strings.")
        resolved[key] = value
    return resolved


def _decode_string_dict(raw: Any, *, field_name: str) -> dict[str, str]:
    if not isinstance(raw, Mapping):
        raise EventCursorDecodeError(f"{field_name} must be a mapping.")
    resolved: dict[str, str] = {}
    for key, value in raw.items():
        if not isinstance(key, str) or not key.strip():
            raise EventCursorDecodeError(f"{field_name} keys must be non-empty strings.")
        if not isinstance(value, str):
            raise EventCursorDecodeError(f"{field_name}[{key!r}] must be a string.")
        resolved[key] = value
    return resolved


def _assert_exact_keys(
    mapping: Mapping[str, Any],
    expected_keys: frozenset[str],
    *,
    field_name: str,
) -> None:
    keys = set(mapping.keys())
    unknown = sorted(key for key in keys - expected_keys)
    if unknown:
        raise EventCursorDecodeError(f"{field_name} contains unknown keys: {unknown}")
    missing = sorted(key for key in expected_keys - keys)
    if missing:
        raise EventCursorDecodeError(f"{field_name} is missing required keys: {missing}")


def _require_mapping(raw: Any, *, field_name: str) -> Mapping[str, Any]:
    if not isinstance(raw, Mapping):
        raise EventCursorDecodeError(f"{field_name} must be a mapping.")
    return raw


def _require_non_empty_str(raw: Any, *, field_name: str) -> str:
    if not isinstance(raw, str):
        raise EventCursorDecodeError(f"{field_name} must be a string.")
    resolved = raw.strip()
    if not resolved:
        raise EventCursorDecodeError(f"{field_name} must not be empty.")
    return resolved


def _require_non_negative_int(raw: Any, *, field_name: str) -> int:
    if isinstance(raw, bool):
        raise EventCursorDecodeError(f"{field_name} must be an integer.")
    if isinstance(raw, int):
        resolved = raw
    elif isinstance(raw, str):
        try:
            resolved = int(raw)
        except ValueError as exc:
            raise EventCursorDecodeError(f"{field_name} must be an integer.") from exc
    else:
        raise EventCursorDecodeError(f"{field_name} must be an integer.")
    if resolved < 0:
        raise EventCursorDecodeError(f"{field_name} must be >= 0.")
    return resolved


__all__ = [
    "EventCursorDecodeError",
    "decode_event_cursor",
    "encode_event_cursor",
]



from __future__ import annotations

from dataclasses import dataclass, field
from typing import Any, Literal, Optional

from flownet.core.types import PayloadData


ExceptionAction = Literal["propagate", "abort", "fallback"]


@dataclass
class ExceptionContext:
    request_id: str
    phase: str
    actor_id: Optional[str] = None
    method: Optional[str] = None
    address: Optional[str] = None
    transformation_id: Optional[str] = None
    payload: Optional[PayloadData] = None
    stack_depth: int = 0
    metadata: dict[str, Any] = field(default_factory=dict)


@dataclass
class ExceptionEvent:
    error_type: str
    message: str
    traceback: str
    context: ExceptionContext
    error: Optional[BaseException] = None


@dataclass
class ExceptionDecision:
    action: ExceptionAction
    payloads: Optional[list[PayloadData]] = None
    metadata: dict[str, Any] = field(default_factory=dict)

    @staticmethod
    def abort(**metadata: Any) -> "ExceptionDecision":
        return ExceptionDecision(action="abort", payloads=None, metadata=metadata)

    @staticmethod
    def propagate(**metadata: Any) -> "ExceptionDecision":
        return ExceptionDecision(action="propagate", payloads=None, metadata=metadata)

    @staticmethod
    def fallback(
        value: Optional[Any] = None,
        payloads: Optional[list[PayloadData]] = None,
        **metadata: Any,
    ) -> "ExceptionDecision":
        resolved_payloads: Optional[list[PayloadData]]
        if payloads is not None:
            resolved_payloads = payloads
        elif value is None:
            resolved_payloads = []
        elif isinstance(value, PayloadData):
            resolved_payloads = [value]
        else:
            resolved_payloads = [PayloadData(args=[value], kwargs={})]
        return ExceptionDecision(action="fallback", payloads=resolved_payloads, metadata=metadata)


class FlowException(RuntimeError):
    def __init__(self, event: ExceptionEvent):
        message = f"{event.error_type}: {event.message}"
        super().__init__(message)
        self.event = event


class FlowDefinitionError(RuntimeError):
    """Raised when a flow definition violates return/sink semantics."""



from __future__ import annotations

from dataclasses import dataclass
import hashlib
from typing import Any, Iterable

from flownet.v1.runtime.flowengine.cursor_models import FlowProgramRef

SCOPE_ROOT_ID = "scope_root"


@dataclass(frozen=True)
class ExceptionScope:
    scope_id: str
    parent_scope_id: str | None
    handler_target: Any
    depth: int


@dataclass(frozen=True)
class CompiledExceptionScopes:
    root_scope_id: str
    scopes: dict[str, ExceptionScope]
    node_scope_ids: dict[str, str]

    def scope_id_for_node(self, pc_node_id: str) -> str:
        scope_id = self.node_scope_ids.get(pc_node_id)
        if scope_id is None:
            raise KeyError(f"unknown_pc_node_id:{pc_node_id}")
        return scope_id

    def handler_target(self, scope_id: str) -> Any:
        scope = self.scopes.get(scope_id)
        if scope is None:
            raise KeyError(f"unknown_scope_id:{scope_id}")
        return scope.handler_target

    def parent_scope_id(self, scope_id: str) -> str | None:
        scope = self.scopes.get(scope_id)
        if scope is None:
            raise KeyError(f"unknown_scope_id:{scope_id}")
        return scope.parent_scope_id

    def iter_scope_chain(self, start_scope_id: str) -> Iterable[str]:
        seen: set[str] = set()
        scope_id: str | None = start_scope_id
        while scope_id is not None:
            if scope_id in seen:
                raise RuntimeError(f"exception_scope_cycle_detected:{scope_id}")
            seen.add(scope_id)
            yield scope_id
            scope_id = self.parent_scope_id(scope_id)


class ProgramScopeRegistry:
    """Resolve compiled exception scopes by `FlowProgramRef`."""

    def __init__(self):
        self._scopes_by_program_ref: dict[tuple[str, str], CompiledExceptionScopes] = {}

    @staticmethod
    def _key(program_ref: FlowProgramRef) -> tuple[str, str]:
        return (program_ref.program_uri, program_ref.program_rev)

    def register(self, program_ref: FlowProgramRef, scopes: CompiledExceptionScopes) -> None:
        self._scopes_by_program_ref[self._key(program_ref)] = scopes

    def resolve(self, program_ref: FlowProgramRef) -> CompiledExceptionScopes:
        scopes = self._scopes_by_program_ref.get(self._key(program_ref))
        if scopes is None:
            raise KeyError(
                "exception_scopes_not_found:"
                f" program_uri={program_ref.program_uri},"
                f" program_rev={program_ref.program_rev}"
            )
        return scopes


def compile_exception_scopes(flow_program: Any) -> CompiledExceptionScopes:
    transformations = _resolve_transformations(flow_program)
    scopes: dict[str, ExceptionScope] = {
        SCOPE_ROOT_ID: ExceptionScope(
            scope_id=SCOPE_ROOT_ID,
            parent_scope_id=None,
            handler_target=None,
            depth=0,
        )
    }
    node_scope_ids: dict[str, str] = {}
    key_to_scope_id: dict[tuple[str, ...], str] = {(): SCOPE_ROOT_ID}

    for transformation in transformations:
        trans_id = _require_transformation_id(transformation)
        handler_stack = _resolve_handler_stack(transformation)

        current_key: tuple[str, ...] = ()
        current_scope_id = SCOPE_ROOT_ID
        for layer in handler_stack:
            layer_key = _layer_key(layer)
            next_key = current_key + (layer_key,)
            scope_id = key_to_scope_id.get(next_key)
            if scope_id is None:
                scope_id = _scope_id_for_key(next_key)
                scopes[scope_id] = ExceptionScope(
                    scope_id=scope_id,
                    parent_scope_id=current_scope_id,
                    handler_target=layer,
                    depth=len(next_key),
                )
                key_to_scope_id[next_key] = scope_id
            current_scope_id = scope_id
            current_key = next_key

        node_scope_ids[trans_id] = current_scope_id

    return CompiledExceptionScopes(
        root_scope_id=SCOPE_ROOT_ID,
        scopes=scopes,
        node_scope_ids=node_scope_ids,
    )


def _resolve_transformations(flow_program: Any) -> list[Any]:
    transformations_attr = getattr(flow_program, "transformations", None)
    if isinstance(transformations_attr, dict) and transformations_attr:
        return list(transformations_attr.values())

    pipeline_attr = getattr(flow_program, "pipeline", None)
    if isinstance(pipeline_attr, list) and pipeline_attr:
        return list(pipeline_attr)
    return []


def _require_transformation_id(transformation: Any) -> str:
    trans_id = getattr(transformation, "trans_id", None)
    if not isinstance(trans_id, str) or not trans_id.strip():
        raise ValueError("transformation.trans_id must be a non-empty string.")
    return trans_id


def _resolve_handler_stack(transformation: Any) -> list[Any]:
    raw_stack = getattr(transformation, "exception_handler_stack", None)
    if raw_stack is None:
        return []
    if not isinstance(raw_stack, list):
        raise ValueError(
            "transformation.exception_handler_stack must be a list when provided.",
        )
    if raw_stack and raw_stack[0] is None:
        return list(raw_stack[1:])
    return list(raw_stack)


def _scope_id_for_key(key: tuple[str, ...]) -> str:
    digest = hashlib.sha1("||".join(key).encode("utf-8")).hexdigest()
    return f"scope_{digest[:16]}"


def _layer_key(handler: Any) -> str:
    if handler is None:
        return "none"
    address = getattr(handler, "address", None)
    actor_id = getattr(handler, "actor_id", None)
    method = getattr(handler, "method", None)
    if not all(isinstance(value, str) and value.strip() for value in (address, actor_id, method)):
        raise TypeError(
            "exception handler layer must be None or object with non-empty "
            "address/actor_id/method strings.",
        )
    return f"actor:{address}:{actor_id}:{method}"


__all__ = [
    "SCOPE_ROOT_ID",
    "ExceptionScope",
    "CompiledExceptionScopes",
    "ProgramScopeRegistry",
    "compile_exception_scopes",
]



from __future__ import annotations

import asyncio
import contextlib
import inspect
from concurrent.futures import Future as ConcurrentFuture
from dataclasses import dataclass, field
from typing import Callable
from typing import Any, Literal

from flownet.v1.runtime.actors.error_codes import (
    ACTOR_TASK_CANCELLED,
    ACTOR_TASK_TIMEOUT,
    build_error_message,
)
from flownet.v1.runtime.actors.event_emitter import (
    ActorEmittedEvent,
    ActorTaskEventBuffer,
    actor_task_event_scope,
    resolve_task_event_policy,
)


SYNC_RETURN_MODE = "sync_return"
TASK_RETURN_MODE = "task_return"
TaskReturnMode = Literal["sync_return", "task_return"]
EventDispatcher = Callable[[ActorEmittedEvent], Any]


@dataclass(frozen=True)
class ActorTaskResult:
    """Resolved actor main-entry result under task-return protocol."""

    result: Any
    return_mode: TaskReturnMode
    event_delivery_reports: tuple[dict[str, Any], ...] = ()
    runtime_report: dict[str, Any] = field(default_factory=dict)


def is_runtime_task_submission(value: Any) -> bool:
    """Return True when actor main-entry returned a runtime task object."""

    if inspect.isgenerator(value):
        return False
    if inspect.isasyncgen(value):
        return False
    if asyncio.iscoroutine(value):
        return True
    if isinstance(value, (asyncio.Future, ConcurrentFuture)):
        return True
    return inspect.isawaitable(value)


class ActorTaskRuntime:
    """Resolve runtime-managed task objects returned by actor main-entry."""

    def __init__(self, *, event_dispatcher: EventDispatcher | None = None) -> None:
        self._event_dispatcher = event_dispatcher

    def set_event_dispatcher(self, event_dispatcher: EventDispatcher | None) -> None:
        self._event_dispatcher = event_dispatcher

    @contextlib.contextmanager
    def open_task_event_scope(
        self,
        *,
        task_policy: Any | None = None,
    ):
        policy = resolve_task_event_policy(task_policy)
        with actor_task_event_scope(policy=policy) as buffer:
            yield buffer

    async def resolve_main_entry_return(
        self,
        value: Any,
        *,
        event_buffer: ActorTaskEventBuffer | None = None,
        task_timeout_ms: int | None = None,
        max_parallel_tools: int | None = None,
        max_pending: int | None = None,
    ) -> ActorTaskResult:
        runtime_report = _build_runtime_report(
            task_timeout_ms=task_timeout_ms,
            max_parallel_tools=max_parallel_tools,
            max_pending=max_pending,
        )
        if not is_runtime_task_submission(value):
            reports = await self._dispatch_events(event_buffer)
            return ActorTaskResult(
                result=value,
                return_mode=SYNC_RETURN_MODE,
                event_delivery_reports=reports,
                runtime_report=runtime_report,
            )
        try:
            resolved_result = await _await_runtime_task(
                value,
                timeout_ms=task_timeout_ms,
            )
        except RuntimeError as exc:
            if str(exc).startswith(ACTOR_TASK_TIMEOUT):
                runtime_report["task_timeout_hit"] = True
            if str(exc).startswith(ACTOR_TASK_CANCELLED):
                runtime_report["task_cancelled"] = True
            await self._dispatch_events(event_buffer)
            raise
        except Exception:
            await self._dispatch_events(event_buffer)
            raise
        reports = await self._dispatch_events(event_buffer)
        return ActorTaskResult(
            result=resolved_result,
            return_mode=TASK_RETURN_MODE,
            event_delivery_reports=reports,
            runtime_report=runtime_report,
        )

    async def _dispatch_events(
        self,
        event_buffer: ActorTaskEventBuffer | None,
    ) -> tuple[dict[str, Any], ...]:
        if event_buffer is None:
            return ()
        events = event_buffer.drain()
        if not events:
            return ()
        reports: list[dict[str, Any]] = []
        for event in events:
            if self._event_dispatcher is None:
                reports.append(
                    {
                        "status": "dropped_no_dispatcher",
                        "mode": event.mode,
                        "topic_uri": event.topic_uri,
                        "event_group_id": event.event_group_id,
                        "seq": event.seq,
                    }
                )
                continue
            try:
                dispatch_result = self._event_dispatcher(event)
                if inspect.isawaitable(dispatch_result):
                    dispatch_result = await dispatch_result
                reports.append(
                    {
                        "status": "dispatched",
                        "mode": event.mode,
                        "topic_uri": event.topic_uri,
                        "event_group_id": event.event_group_id,
                        "seq": event.seq,
                        "dispatch_result": dispatch_result,
                    }
                )
            except Exception as exc:
                reports.append(
                    {
                        "status": "dispatch_error",
                        "mode": event.mode,
                        "topic_uri": event.topic_uri,
                        "event_group_id": event.event_group_id,
                        "seq": event.seq,
                        "error_type": exc.__class__.__name__,
                        "message": str(exc),
                    }
                )
        return tuple(reports)


async def _await_runtime_task(value: Any, *, timeout_ms: int | None = None) -> Any:
    timeout_seconds: float | None = None
    if timeout_ms is not None:
        timeout_seconds = float(timeout_ms) / 1000.0
    awaitable = _coerce_runtime_task_awaitable(value)
    if timeout_seconds is None:
        return await awaitable
    try:
        return await asyncio.wait_for(awaitable, timeout=timeout_seconds)
    except asyncio.TimeoutError as exc:
        raise RuntimeError(
            build_error_message(
                ACTOR_TASK_TIMEOUT,
                task_timeout_ms=int(timeout_ms),
            )
        ) from exc
    except asyncio.CancelledError as exc:
        raise RuntimeError(ACTOR_TASK_CANCELLED) from exc


def _coerce_runtime_task_awaitable(value: Any):
    if asyncio.iscoroutine(value):
        return value
    if isinstance(value, asyncio.Future):
        return value
    if isinstance(value, ConcurrentFuture):
        return asyncio.wrap_future(value)
    if inspect.isawaitable(value):
        return value
    raise TypeError(
        "runtime_task_submission must be coroutine, awaitable, "
        "asyncio.Future, or concurrent.futures.Future.",
    )


def _build_runtime_report(
    *,
    task_timeout_ms: int | None,
    max_parallel_tools: int | None,
    max_pending: int | None,
) -> dict[str, Any]:
    report: dict[str, Any] = {
        "task_timeout_hit": False,
        "task_cancelled": False,
    }
    if task_timeout_ms is not None:
        report["task_timeout_ms"] = int(task_timeout_ms)
    if max_parallel_tools is not None:
        report["max_parallel_tools"] = int(max_parallel_tools)
    if max_pending is not None:
        report["max_pending"] = int(max_pending)
    return report


__all__ = [
    "SYNC_RETURN_MODE",
    "TASK_RETURN_MODE",
    "TaskReturnMode",
    "EventDispatcher",
    "ActorTaskResult",
    "is_runtime_task_submission",
    "ActorTaskRuntime",
]



from __future__ import annotations

import queue
import threading
import uuid
from typing import TYPE_CHECKING, Any, Callable

from flownet.api import flow
from flownet.core.lockfree import lockfree
from flownet.core.request import FlowAckRoute
from flownet.core.types import PayloadData
from flownet.runtime.contexts.flow import current_flow_request
from flownet.runtime.flows.ordered_event_channel import (
    EventEnvelope,
    GapNotice,
    OrderedEventChannel,
    TerminalEnvelope,
)

if TYPE_CHECKING:
    from flownet.api.actor_method_ref import ActorMethodRef
    from flownet.api.datastream import DataStream


class RunOutputRequest:
    """Request-scoped helper for grouped writes on a shared request_id."""

    def __init__(self, channel: "RunIOChannel", request_id: str | None = None):
        self._channel = channel
        self.request_id = request_id
        self._finished = False

    def write(self, *args, **kwargs) -> str:
        request_id = self.request_id
        if request_id is None:
            request_id = self._channel.write(*args, autocommit=False, **kwargs)
            self.request_id = request_id
            return request_id
        self._channel.write(*args, request_id=request_id, autocommit=False, **kwargs)
        return request_id

    def finish(self) -> bool:
        request_id = self.request_id
        if not request_id:
            return False
        self._finished = self._channel.finish(request_id)
        return self._finished

    def collect(
        self,
        *,
        timeout: float = 5.0,
        late_drain_timeout: float = 1.0,
    ) -> list[Any]:
        request_id = self.request_id
        if not request_id:
            return []
        return self._channel.collect(
            request_id=request_id,
            timeout=timeout,
            late_drain_timeout=late_drain_timeout,
        )

    def __enter__(self) -> "RunOutputRequest":
        return self

    def __exit__(self, exc_type, exc, tb) -> bool:
        if exc_type is None:
            self.finish()
        return False


class RunIOChannel:
    """
    Runtime-managed compatibility output adapter used by buffered/return_once runs.
    This class is intentionally transport-focused: it should not own run lifecycle
    semantics, which belong to FlowRunRef.
    """

    def __init__(
        self,
        target: "ActorMethodRef",
        *,
        runtime_managed: bool = True,
        ordered_capacity: int = 256,
        ordered_backpressure: str = "block",
        ordered_terminal_hold_s: float = 0.05,
    ):
        self.target = target
        self.runtime_managed = runtime_managed

        self.return_buffer = queue.Queue()
        self._ordered_capacity = int(ordered_capacity)
        self._ordered_backpressure = str(ordered_backpressure)
        self._ordered_terminal_hold_s = float(ordered_terminal_hold_s)
        self._request_channels: dict[str, OrderedEventChannel] = {}
        self._buffers_lock = threading.Lock()
        self._pending_requests: dict[str, int] = {}
        self._pending_lock = threading.Lock()
        self._actor_id: str | None = None
        self._actor_ref = None
        self._feed_ref = None
        self._runtime_start_request: Callable[[str, int], Any] | None = None
        self._runtime_ack_request_delta: Callable[[str, int], Any] | None = None
        self._runtime_next_request_meta: Callable[[str], dict[str, Any]] | None = None
        self._runtime_finish_request: Callable[[str], Any] | None = None
        self._runtime_close_ingress: Callable[[], bool] | None = None
        self._runtime_request_done: Callable[[str], bool] | None = None

    @flow
    def output_task(self, init_stream: "DataStream"):
        return init_stream.pipe(self.target).map(self._get_feed_ref())

    def feed(self, *args, **kwargs):
        request = current_flow_request()
        request_id = request.request_id if request is not None else None
        data = PayloadData(args=args, kwargs=kwargs)
        if request_id:
            channel = self._get_request_channel(request_id, create=False)
            if channel is None and self._has_pending_request(request_id):
                channel = self._get_request_channel(request_id, create=True)
            if channel is not None:
                channel.publish_data(data, kind="data")
        self.return_buffer.put(data)

    def _get_request_channel(
        self,
        request_id: str,
        create: bool = True,
    ) -> OrderedEventChannel | None:
        if not request_id:
            return None
        with self._buffers_lock:
            channel = self._request_channels.get(request_id)
            if channel is None and create:
                channel = OrderedEventChannel(
                    request_id=request_id,
                    capacity=self._ordered_capacity,
                    backpressure=self._ordered_backpressure,  # type: ignore[arg-type]
                    terminal_hold_s=self._ordered_terminal_hold_s,
                )
                self._request_channels[request_id] = channel
            return channel

    def _release_request_channel(self, request_id: str):
        if not request_id:
            return
        channel: OrderedEventChannel | None
        with self._buffers_lock:
            channel = self._request_channels.pop(request_id, None)
        if channel is not None:
            channel.close()

    def _has_pending_request(self, request_id: str) -> bool:
        with self._pending_lock:
            return request_id in self._pending_requests

    def _get_actor_id(self) -> str:
        if self._actor_id is None:
            from flownet.api.runtime import make_object_actor

            self._actor_ref = make_object_actor(self)
            self._actor_id = self._actor_ref.actor_id
        return self._actor_id

    def _get_actor_address(self) -> str | None:
        actor_ref = self._actor_ref
        if actor_ref is None:
            return None
        normalized = str(getattr(actor_ref, "address", "") or "").strip()
        return normalized or None

    def _get_feed_ref(self):
        if self._feed_ref is None:
            from flownet.api.actor_method_ref import ActorMethodRef

            actor_ref = self._actor_ref
            if actor_ref is None:
                from flownet.api.runtime import make_object_actor

                actor_ref = make_object_actor(self)
                self._actor_ref = actor_ref
                self._actor_id = actor_ref.actor_id
            self._feed_ref = ActorMethodRef(
                address=actor_ref.address,
                actor_id=actor_ref.actor_id,
                method="feed",
            )
        return self._feed_ref

    def ack_route(self) -> FlowAckRoute:
        return FlowAckRoute(
            coordinator_target=self._get_actor_id(),
            coordinator_address=self._get_actor_address(),
        )

    @lockfree
    def read(self, timeout: float = 5.0, request_id: str | None = None):
        if request_id is None:
            result = self.return_buffer.get(timeout=timeout)
            if isinstance(result, (RunOutputDone, RunOutputError)):
                return result
            return result.unpack()

        channel = self._get_request_channel(request_id, create=True)
        assert channel is not None
        result = channel.read(timeout=timeout)
        if result is None:
            if not self._has_pending_request(request_id):
                self._release_request_channel(request_id)
            raise queue.Empty()
        decoded = self._decode_channel_item(request_id, result)
        if isinstance(decoded, RunOutputError):
            self._release_request_channel(request_id)
        return decoded

    def write(
        self,
        *args,
        request_id: str | None = None,
        autocommit: bool | None = None,
        **kwargs,
    ):
        new_request = request_id is None
        if new_request:
            request_id = f"{self._get_actor_id()}:{uuid.uuid4().hex}"
        elif not isinstance(request_id, str) or not request_id:
            raise ValueError("request_id must be a non-empty string when provided.")
        elif not self._has_pending_request(request_id):
            raise ValueError(
                "Custom runtime request_id injection is not allowed. "
                "Call write() without request_id to start a request and reuse the returned id.",
            )

        assert request_id is not None
        with self._pending_lock:
            self._pending_requests[request_id] = self._pending_requests.get(request_id, 0) + 1
        self._get_request_channel(request_id, create=True)
        if self._runtime_start_request is not None:
            self._runtime_start_request(request_id, 1)
        next_meta: dict[str, Any] | None = None
        if self._runtime_next_request_meta is not None:
            next_meta = self._runtime_next_request_meta(request_id)
        dispatched_request_id = self.output_task.put_request(
            request_id,
            *args,
            ack_route=self.ack_route(),
            _next_meta=next_meta,
            **kwargs,
        )
        if autocommit is None:
            autocommit = new_request
        if autocommit:
            self.finish(request_id)
        return dispatched_request_id

    def run(
        self,
        *args,
        collect_timeout: float = 5.0,
        late_drain_timeout: float = 1.0,
        **kwargs,
    ) -> list[Any]:
        request_id = self.write(*args, autocommit=True, **kwargs)
        return self.collect(
            request_id=request_id,
            timeout=collect_timeout,
            late_drain_timeout=late_drain_timeout,
        )

    def finish(self, request_id: str) -> bool:
        if not request_id:
            return False
        if self._runtime_finish_request is not None:
            finished = bool(self._runtime_finish_request(request_id))
            if finished:
                self.event_chain_done(request_id, 0)
            return finished
        return False

    def collect(
        self,
        *,
        request_id: str,
        timeout: float = 5.0,
        late_drain_timeout: float = 1.0,
    ) -> list[Any]:
        if not request_id:
            raise ValueError("request_id is required for collect().")
        outputs: list[Any] = []
        done_seen = False
        while True:
            try:
                item = self.read(
                    timeout=late_drain_timeout if done_seen else timeout,
                    request_id=request_id,
                )
            except queue.Empty:
                if done_seen:
                    break
                raise
            if isinstance(item, RunOutputDone):
                done_seen = True
                continue
            if isinstance(item, RunOutputError):
                raise RuntimeError(
                    f"{item.error_type}: {item.message}\n{item.traceback_text}"
                )
            if isinstance(item, RunOutputGap):
                continue
            outputs.append(item)
        return outputs

    def close(self) -> bool:
        closed = False
        if self._runtime_close_ingress is not None:
            closed = bool(self._runtime_close_ingress())
        if closed:
            with self._pending_lock:
                pending_ids = tuple(self._pending_requests.keys())
            for request_id in pending_ids:
                self.event_chain_done(request_id, 0)
        channels: list[OrderedEventChannel]
        with self._buffers_lock:
            channels = list(self._request_channels.values())
            self._request_channels.clear()
        for channel in channels:
            channel.close()
        return closed

    def close_ingress(self) -> bool:
        return self.close()

    def request(self, request_id: str | None = None) -> RunOutputRequest:
        return RunOutputRequest(self, request_id=request_id)

    def __enter__(self) -> "RunIOChannel":
        return self

    def __exit__(self, exc_type, exc, tb) -> bool:
        self.close()
        return False

    def subscribe(
        self,
        handler: Callable[[Any], Any],
        *,
        request_id: str,
        from_seq: int | None = None,
        poll_timeout: float = 0.2,
    ):
        if not request_id:
            raise ValueError("request_id is required for subscribe().")
        channel = self._get_request_channel(request_id, create=True)
        assert channel is not None

        def _on_item(item: EventEnvelope | TerminalEnvelope | GapNotice):
            handler(self._decode_channel_item(request_id, item))

        return channel.subscribe(
            _on_item,
            from_seq=from_seq,
            poll_timeout=poll_timeout,
        )

    def request_delivery_done(self, request_id: str) -> bool:
        channel = self._get_request_channel(request_id, create=False)
        if channel is None:
            return not self._has_pending_request(request_id)
        return self._channel_delivery_done(channel)

    def request_delivery_status(self, request_id: str) -> dict[str, Any]:
        channel = self._get_request_channel(request_id, create=False)
        if channel is None:
            return {
                "request_id": request_id,
                "request_ref_id": request_id,
                "delivery_done": not self._has_pending_request(request_id),
                "watermark_done": not self._has_pending_request(request_id),
                "terminal_emitted": False,
                "drained_seq": 0,
                "final_seq": None,
                "stats": {},
            }
        stats = channel.stats()
        terminal_final_seq = channel.terminal_final_seq()
        next_read_seq = int(stats.get("next_read_seq", 1))
        terminal_emitted = channel.terminal_emitted()
        drained_seq = max(0, next_read_seq - 1)
        watermark_done = (
            terminal_final_seq is not None
            and terminal_emitted
            and next_read_seq > terminal_final_seq
        )
        return {
            "request_id": request_id,
            "request_ref_id": request_id,
            "delivery_done": watermark_done,
            "watermark_done": watermark_done,
            "terminal_emitted": terminal_emitted,
            "drained_seq": drained_seq,
            "final_seq": terminal_final_seq,
            "stats": stats,
        }

    @staticmethod
    def _channel_delivery_done(channel: OrderedEventChannel) -> bool:
        terminal_final_seq = channel.terminal_final_seq()
        if terminal_final_seq is None:
            return False
        if not channel.terminal_emitted():
            return False
        stats = channel.stats()
        try:
            next_read_seq = int(stats.get("next_read_seq", 1))
        except (TypeError, ValueError):
            next_read_seq = 1
        return next_read_seq > terminal_final_seq

    def event_chain_done(self, request_id: str, delta: int):
        runtime_pending: int | None = None
        if self._runtime_ack_request_delta is not None:
            runtime_pending = self._runtime_ack_request_delta(request_id, delta)
        done = False
        with self._pending_lock:
            pending = self._pending_requests.get(request_id)
            if pending is None and runtime_pending is None:
                return
            pending = (pending or 0) + delta
            if runtime_pending is not None:
                pending = int(runtime_pending)
            if pending <= 0:
                request_done = True
                if self._runtime_request_done is not None:
                    request_done = bool(self._runtime_request_done(request_id))
                if request_done:
                    self._pending_requests.pop(request_id, None)
                    done = True
                else:
                    # Keep request open until producer_done arrives.
                    self._pending_requests[request_id] = 0
            else:
                self._pending_requests[request_id] = pending
        if done:
            channel = self._get_request_channel(request_id, create=True)
            assert channel is not None
            final_seq = channel.last_published_seq()
            channel.publish_terminal(final_seq=final_seq, reason="runtime_done")
            done_event = RunOutputDone(
                request_id,
                final_seq=final_seq,
                expected_total_events=final_seq,
                reason="runtime_done",
                request_done=True,
                status="completed",
                subscriber_done=True,
            )
            self.return_buffer.put(done_event)
            return

    def flow_error(
        self,
        request_id: str,
        error_type: str,
        message: str,
        traceback_text: str,
        event: Any | None = None,
    ):
        error = RunOutputError(
            request_id=request_id,
            error_type=error_type,
            message=message,
            traceback_text=traceback_text,
            event=event,
        )
        if request_id:
            with self._pending_lock:
                pending = self._pending_requests.pop(request_id, None)
            if pending is not None and self._runtime_ack_request_delta is not None:
                self._runtime_ack_request_delta(request_id, -pending)
            channel = self._get_request_channel(request_id, create=True)
            if channel is not None:
                channel.publish_data(error, kind="system")
                final_seq = channel.last_published_seq()
                channel.publish_terminal(final_seq=final_seq, reason="runtime_error")
        self.return_buffer.put(error)

    def _decode_channel_item(
        self,
        request_id: str,
        item: EventEnvelope | TerminalEnvelope | GapNotice,
    ) -> Any:
        if isinstance(item, GapNotice):
            return RunOutputGap(
                request_id=request_id,
                seq=item.seq,
                reason=item.reason,
            )
        if isinstance(item, TerminalEnvelope):
            status = "failed" if str(item.reason or "").strip().lower() == "runtime_error" else "completed"
            return RunOutputDone(
                request_id=request_id,
                final_seq=item.final_seq,
                expected_total_events=item.final_seq,
                reason=item.reason,
                request_done=True,
                status=status,
                subscriber_done=True,
            )
        payload = item.payload
        if isinstance(payload, PayloadData):
            return payload.unpack()
        return payload

    def bind_runtime_tracker(
        self,
        *,
        start_request: Callable[[str, int], Any] | None = None,
        ack_request_delta: Callable[[str, int], Any] | None = None,
        next_request_meta: Callable[[str], dict[str, Any]] | None = None,
        finish_request: Callable[[str], Any] | None = None,
        close_ingress: Callable[[], bool] | None = None,
        request_done: Callable[[str], bool] | None = None,
    ) -> None:
        self._runtime_start_request = start_request
        self._runtime_ack_request_delta = ack_request_delta
        self._runtime_next_request_meta = next_request_meta
        self._runtime_finish_request = finish_request
        self._runtime_close_ingress = close_ingress
        self._runtime_request_done = request_done


class RunOutputError:
    def __init__(
        self,
        request_id: str,
        error_type: str,
        message: str,
        traceback_text: str,
        event: Any | None = None,
    ):
        self.request_id = request_id
        self.error_type = error_type
        self.message = message
        self.traceback_text = traceback_text
        self.event = event

    @property
    def request_ref_id(self) -> str:
        """Compatibility alias of request_id for request-ref terminology."""
        return self.request_id


class RunOutputDone:
    def __init__(
        self,
        request_id: str,
        *,
        final_seq: int | None = None,
        expected_total_events: int | None = None,
        reason: str | None = None,
        request_done: bool | None = None,
        status: str | None = None,
        subscriber_done: bool | None = None,
    ):
        self.request_id = request_id
        self.final_seq = final_seq
        self.expected_total_events = expected_total_events
        self.reason = reason
        self.request_done = request_done
        self.status = status
        self.subscriber_done = subscriber_done

    @property
    def request_ref_id(self) -> str:
        """Compatibility alias of request_id for request-ref terminology."""
        return self.request_id


class RunOutputGap:
    def __init__(self, request_id: str, seq: int, reason: str):
        self.request_id = request_id
        self.seq = seq
        self.reason = reason

    @property
    def request_ref_id(self) -> str:
        """Compatibility alias of request_id for request-ref terminology."""
        return self.request_id


# Backward-compatible alias during naming migration from output-only wording
# to explicit run ingress+egress IO semantics.
RunOutputChannel = RunIOChannel


__all__ = [
    "RunIOChannel",
    "RunOutputChannel",
    "RunOutputRequest",
    "RunOutputDone",
    "RunOutputGap",
    "RunOutputError",
]



import argparse
import os
import threading

from flownet.config import NodeRuntimeConfig
from flownet.platform.node_runtime import bootstrap_node_runtime


def main():
    os.environ.setdefault("FLOWNET_RUNTIME_ROLE", "node")
    parser = argparse.ArgumentParser(description="Start a Flownet node runtime.")
    parser.add_argument("--host", type=str, default="127.0.0.1", help="Host to bind the node runtime.")
    parser.add_argument("--port", type=int, required=True, help="Port to bind the node runtime.")
    parser.add_argument("--threads", type=int, default=4, help="Max worker threads for the node runtime.")
    parser.add_argument("--transport-mode", type=str, default="auto", help="Transport mode: auto|tcp|uds_prefer|hetero_auto.")
    parser.add_argument("--local-ipc-dir", type=str, default="/tmp/flownet", help="Directory for local IPC sockets.")
    parser.add_argument(
        "--transport-shm-threshold-bytes",
        type=int,
        default=262144,
        help="Payload size threshold to prefer SHM in hetero_auto mode.",
    )
    parser.add_argument(
        "--transport-shm-max-payload-bytes",
        type=int,
        default=67108864,
        help="Max payload bytes allowed for SHM transport.",
    )
    parser.add_argument(
        "--transport-backend-cooldown-seconds",
        type=float,
        default=1.0,
        help="Cooldown after backend failure before retrying same backend/address.",
    )
    parser.add_argument(
        "--transport-backend-cooldown-max-entries",
        type=int,
        default=4096,
        help="Max entries tracked in backend cooldown map.",
    )
    parser.add_argument(
        "--transport-backend-cooldown-cleanup-interval",
        type=int,
        default=128,
        help="Cleanup interval (send operations) for cooldown pruning.",
    )
    parser.add_argument("--node-id", type=str, default="default-node", help="Logical node id.")
    args = parser.parse_args()

    config = NodeRuntimeConfig(
        host=args.host,
        port=args.port,
        threads=args.threads,
        node_id=args.node_id,
        transport_mode=args.transport_mode,
        local_ipc_dir=args.local_ipc_dir,
        transport_shm_threshold_bytes=args.transport_shm_threshold_bytes,
        transport_shm_max_payload_bytes=args.transport_shm_max_payload_bytes,
        transport_backend_cooldown_seconds=args.transport_backend_cooldown_seconds,
        transport_backend_cooldown_max_entries=args.transport_backend_cooldown_max_entries,
        transport_backend_cooldown_cleanup_interval=args.transport_backend_cooldown_cleanup_interval,
    )
    bootstrap_node_runtime(config, node_id=args.node_id)
    print(f"[NodeRuntime Main] FlownetRuntime started at {args.host}:{args.port} with {args.threads} threads.")

    try:
        threading.Event().wait()
    except KeyboardInterrupt:
        print("[NodeRuntime Main] Shutdown requested. Exiting...")


if __name__ == "__main__":
    main()



from __future__ import annotations

from dataclasses import dataclass
import inspect
import threading
import uuid
from typing import Any, Callable

from flownet.core.stream_event import StreamEvent
from flownet.runtime.contexts.actor import _user_execution_scope
from flownet.runtime.contexts.source import (
    StreamSourceContext,
    stream_source_context,
    emit,
    _stream_source_context,
)
from flownet.runtime.loops import submit_runtime_coro
from flownet.runtime.runner_lifecycle import start_runner_lifecycle
from flownet.runtime.topics.topic_registry import TopicRegistry


 


@dataclass
class StreamSourceHandle:
    source_id: str
    source: Any
    status: str = "running"
    restart_policy: str = "none"
    last_error: Exception | None = None
    runner: "StreamSourceRunner" | None = None


class StreamSourceRunner:
    def __init__(
        self,
        source_id: str,
        stop_event: threading.Event,
        thread: threading.Thread | None = None,
        future: Any | None = None,
        close_fn: Callable[[], None] | None = None,
    ):
        self._source_id = source_id
        self._stop_event = stop_event
        self._thread = thread
        self._future = future
        self._close_fn = close_fn

    def stop(self) -> None:
        self._stop_event.set()
        if self._close_fn is not None:
            try:
                self._close_fn()
            except Exception as exc:  # pragma: no cover - best effort cleanup
                print(f"[StreamSourceRunner] Failed to close {self._source_id}: {exc}")
        if self._future is not None:
            try:
                self._future.cancel()
            except Exception as exc:  # pragma: no cover - best effort
                print(f"[StreamSourceRunner] Failed to cancel {self._source_id}: {exc}")
        if self._thread is not None and self._thread.is_alive():
            self._thread.join(timeout=1.0)


def _emit_item(context: StreamSourceContext, item: Any) -> None:
    if item is None:
        return
    if isinstance(item, StreamEvent):
        context.emit(item.event_type, item.payload, dict(item.tags))
        return
    if isinstance(item, tuple):
        if len(item) == 2:
            event_type, payload = item
            tags = None
        elif len(item) == 3:
            event_type, payload, tags = item
        else:
            raise ValueError(f"Invalid stream event tuple: {item}")
        context.emit(event_type, payload, tags)
        return
    context.emit("event", item, None)


def _emit_done(context: StreamSourceContext) -> None:
    context.emit("__done__", None, {"__control__": "done"})


def _call_source(
    source_spec: Any,
    *args,
    **kwargs,
) -> tuple[Any, Any]:
    if inspect.isclass(source_spec):
        source_obj = source_spec(*args, **kwargs)
        result = source_obj()
        return source_obj, result
    if inspect.isgenerator(source_spec) or inspect.isasyncgen(source_spec):
        return source_spec, source_spec
    if not callable(source_spec):
        raise TypeError(f"Stream source spec is not callable: {type(source_spec)}")
    result = source_spec(*args, **kwargs)
    return source_spec, result


class StreamSourceRegistry:
    def __init__(
        self,
        call_manager: Any,
        topic_registry: TopicRegistry,
        local_address_fn: Callable[[], str],
        submit_coro: Callable[[Any], Any] = submit_runtime_coro,
        submit_user_coro: Callable[[Any], Any] | None = None,
        request_id_factory: Callable[[str], str] | None = None,
        mark_done_fn: Callable[[str], Any] | None = None,
    ):
        self._call_manager = call_manager
        self._topic_registry = topic_registry
        self._local_address_fn = local_address_fn
        self._submit_coro = submit_coro
        self._submit_user_coro = submit_user_coro or submit_coro
        self._request_id_factory = request_id_factory
        self._mark_done_fn = mark_done_fn
        self._sources: dict[str, StreamSourceHandle] = {}
        self._lock = threading.Lock()

    def register_source(
        self,
        source_id: str,
        source_spec: Any,
        *args,
        restart_policy: str = "none",
        **kwargs,
    ) -> StreamSourceHandle:
        with self._lock:
            if source_id in self._sources:
                raise ValueError(f"Stream source already running: {source_id}")

        stop_event = threading.Event()
        request_id = self._build_request_id(source_id)
        context = StreamSourceContext(
            source_id=source_id,
            registry=self,
            topic_registry=self._topic_registry,
            call_manager=self._call_manager,
            local_address_fn=self._local_address_fn,
            stop_event=stop_event,
            request_id=request_id,
            mark_done_fn=self._mark_done_fn,
        )

        source_obj, result = _call_source(source_spec, *args, **kwargs)

        handle = StreamSourceHandle(
            source_id=source_id,
            source=source_obj,
            restart_policy=restart_policy,
        )
        with self._lock:
            self._sources[source_id] = handle

        try:
            runner = self._start_runner(source_id, context, stop_event, result, handle)
            handle.runner = runner
        except Exception:
            with self._lock:
                self._sources.pop(source_id, None)
            raise
        return handle

    def start_source(
        self,
        source_id: str,
        source_spec: Any,
        *args,
        **kwargs,
    ) -> StreamSourceHandle:
        return self.register_source(source_id, source_spec, *args, **kwargs)

    def _start_runner(
        self,
        source_id: str,
        context: StreamSourceContext,
        stop_event: threading.Event,
        result: Any,
        handle: StreamSourceHandle,
    ) -> StreamSourceRunner | None:
        def _on_stopped(runner_kind: str) -> None:
            if runner_kind in {"asyncgen", "generator"}:
                if not stop_event.is_set():
                    _emit_done(context)
            else:
                _emit_done(context)
            self._update_status(source_id, "stopped", None)

        def _on_cancelled() -> None:
            self._update_status(source_id, "stopped", None)

        def _on_error(exc: Exception) -> None:
            self._update_status(source_id, "error", exc)
            print(f"[StreamSourceRegistry] Source {source_id} failed: {exc}")

        started = start_runner_lifecycle(
            name=source_id,
            result=result,
            stop_event=stop_event,
            submit_user_coro=self._submit_user_coro,
            set_context=lambda: _stream_source_context.set(context),
            reset_context=_stream_source_context.reset,
            execution_scope_factory=lambda: _user_execution_scope(None),
            emit_item=lambda item: _emit_item(context, item),
            on_stopped=_on_stopped,
            on_error=_on_error,
            on_cancelled=_on_cancelled,
            thread_name=f"source-{source_id}",
        )
        if started is None:
            return None
        return StreamSourceRunner(
            source_id=source_id,
            stop_event=stop_event,
            thread=started.thread,
            future=started.future,
            close_fn=started.close_fn,
        )

    def submit(self, coro: Any, *, loop: str = "runtime"):
        if loop == "user" and self._submit_user_coro is not None:
            return self._submit_user_coro(coro)
        return self._submit_coro(coro)

    def _update_status(self, source_id: str, status: str, error: Exception | None) -> None:
        with self._lock:
            handle = self._sources.get(source_id)
            if handle is None:
                return
            handle.status = status
            handle.last_error = error

    def stop_source(self, source_id: str) -> bool:
        with self._lock:
            handle = self._sources.get(source_id)
        if handle is None:
            return False
        if handle.runner is not None:
            handle.runner.stop()
        stop_method = getattr(handle.source, "stop", None)
        if callable(stop_method):
            stop_result = stop_method()
            if inspect.isawaitable(stop_result):
                self._submit_user_coro(stop_result)
        with self._lock:
            self._sources.pop(source_id, None)
        return True

    def list_sources(self) -> list[dict[str, Any]]:
        with self._lock:
            handles = list(self._sources.values())
        return [
            {
                "source_id": handle.source_id,
                "status": handle.status,
                "restart_policy": handle.restart_policy,
                "last_error": str(handle.last_error) if handle.last_error else None,
            }
            for handle in handles
        ]

    def get_source(self, source_id: str) -> StreamSourceHandle | None:
        with self._lock:
            return self._sources.get(source_id)

    def _build_request_id(self, source_id: str) -> str:
        if callable(self._request_id_factory):
            try:
                request_id = self._request_id_factory(source_id)
            except TypeError:
                request_id = self._request_id_factory()  # type: ignore[misc]
            if isinstance(request_id, str) and request_id:
                return request_id
        return uuid.uuid4().hex


__all__ = [
    "StreamSourceContext",
    "StreamSourceHandle",
    "StreamSourceRegistry",
    "StreamSourceRunner",
    "stream_source_context",
    "emit",
]



from __future__ import annotations

from typing import TYPE_CHECKING

from flownet.engine.operators.base import OperatorHandler
from flownet.engine.operators.router import RouterOperatorHandler
from flownet.engine.operators.shuffle import ShuffleRouterOperatorHandler
from flownet.engine.operators.stateless import FilterOperatorHandler, PipeOperatorHandler
from flownet.engine.operators.stateful_process import StatefulProcessOperatorHandler
from flownet.engine.operators.reducers import (
    JoinReducerOperatorHandler,
    MergeReducerOperatorHandler,
    ReducerOperatorHandler,
)

if TYPE_CHECKING:
    from flownet.engine.operator_registry import OperatorRegistry


def register_default_operators(registry: "OperatorRegistry") -> "OperatorRegistry":
    default_handler = registry.default_handler
    registry.register("map", default_handler)
    registry.register("flatmap", default_handler)
    registry.register("tap", default_handler)
    registry.register("filter", FilterOperatorHandler())
    registry.register("pipe", PipeOperatorHandler())
    registry.register_internal("sys/router", RouterOperatorHandler())
    registry.register_internal("sys/shuffle", ShuffleRouterOperatorHandler())
    registry.register_internal("sys/stateful-process", StatefulProcessOperatorHandler())
    registry.register_internal("sys/reducer", ReducerOperatorHandler())
    registry.register_internal("sys/join", ShuffleRouterOperatorHandler())
    registry.register_internal("sys/join-reducer", JoinReducerOperatorHandler())
    registry.register_internal("sys/merge-reducer", MergeReducerOperatorHandler())
    return registry


__all__ = ["register_default_operators"]
